"""Service configuration table model for the service settings editor."""

from .compat import QtCore, QtGui

QAbstractTableModel = QtCore.QAbstractTableModel
QModelIndex = QtCore.QModelIndex
Qt = QtCore.Qt
pyqtSignal = QtCore.pyqtSignal
QColor = QtGui.QColor
QFont = QtGui.QFont

from ..service_settings import SERVICE_SETTINGS, WidgetType


class _ServiceConfigModel(QAbstractTableModel):
    """Table model that holds settings (key/value pairs) for a single service."""

    KEY_COL = 0
    VALUE_COL = 1

    is_dirty_changed = pyqtSignal(bool)

    def __init__(self, service_name: str, service_config: dict):
        super().__init__()
        self._service_name = service_name
        self._model_data = service_config
        self._original_data = service_config.copy()
        self._settings_data = SERVICE_SETTINGS
        self._dirty = False

    def rowCount(self, parent=QModelIndex()):
        return len(self._model_data)

    def columnCount(self, parent=QModelIndex()):
        return 2

    def index_to_setting_key(self, index: QModelIndex) -> str:
        return list(self._model_data.keys())[index.row()]

    def add_settings(self, settings: dict[str, str]):
        self.beginInsertRows(QModelIndex(), self.rowCount(), self.rowCount() + len(settings) - 1)
        self._model_data.update(settings)
        self._set_dirty_status(True)
        self.endInsertRows()

        if self._model_data == self._original_data:
            self._set_dirty_status(False)

    def remove_setting(self, index: QModelIndex):
        if not index.isValid():
            return

        self.beginRemoveRows(QModelIndex(), index.row(), index.row())
        del self._model_data[list(self._model_data.keys())[index.row()]]
        self._set_dirty_status(True)
        self.endRemoveRows()

        if self._model_data == self._original_data:
            self._set_dirty_status(False)

    def data(self, index: QModelIndex, role=Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None

        key = list(self._model_data.keys())[index.row()]

        if role == Qt.ItemDataRole.DisplayRole:
            if index.column() == self.KEY_COL:
                return key
            elif index.column() == self.VALUE_COL:
                setting = self._settings_data.get(key, {})
                if setting.get("widget_type") == WidgetType.PASSWORD:
                    return "************"
                return self._model_data[key]

        elif role == Qt.ItemDataRole.ToolTipRole:
            if index.column() == self.KEY_COL:
                return self._settings_data.get(key, {}).get("description")

        elif role == Qt.ItemDataRole.EditRole and index.column() == self.VALUE_COL:
            return self._model_data[key]

        elif role == Qt.ItemDataRole.FontRole:
            if index.column() == self.KEY_COL:
                font = QFont()
                font.setBold(True)
                return font
            elif index.column() == self.VALUE_COL and (
                key not in self._original_data or self._model_data[key] != self._original_data[key]
            ):
                font = QFont()
                font.setItalic(True)
                return font

        elif role == Qt.ItemDataRole.ForegroundRole and index.column() == self.VALUE_COL:
            if key not in self._original_data or self._model_data[key] != self._original_data[key]:
                return QColor(Qt.GlobalColor.darkGreen)

        elif role == Qt.ItemDataRole.UserRole:
            setting = self._settings_data.get(key, {})
            wtype = setting.get("widget_type")
            if index.column() == self.VALUE_COL and isinstance(wtype, WidgetType):
                return {
                    "widget_type": wtype,
                    "config": setting.get("config"),
                }
            return {}

        return None

    def setData(self, index: QModelIndex, value, role=Qt.ItemDataRole.EditRole) -> bool:
        if not index.isValid():
            return False

        key = list(self._model_data.keys())[index.row()]
        if value != self._model_data[key]:
            self._model_data[key] = value

            if key not in self._original_data or value != self._original_data[key]:
                self._set_dirty_status(True)
            else:
                if self._model_data == self._original_data:
                    self._set_dirty_status(False)

            return True

        return False

    def flags(self, idx: QModelIndex):
        if not idx.isValid():
            return ~Qt.ItemFlag.ItemIsSelectable & ~Qt.ItemFlag.ItemIsEnabled

        _flags = Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEnabled
        if idx.column() == self.VALUE_COL:
            _flags |= Qt.ItemFlag.ItemIsEditable
        return _flags

    def is_dirty(self) -> bool:
        return self._dirty

    def _set_dirty_status(self, status: bool):
        self._dirty = status
        self.is_dirty_changed.emit(status)

    def current_setting_keys(self) -> list[str]:
        return list(self._model_data.keys())

    def service_config(self) -> dict:
        return self._model_data.copy()

    def service_name(self) -> str:
        return self._service_name

    def set_not_dirty(self):
        self._original_data = self._model_data.copy()
        self._set_dirty_status(False)

    def invalid_settings(self) -> list[str]:
        return [k for k, v in self._model_data.items() if v.strip() == ""]

    @staticmethod
    def is_custom_widget_cell(index: QModelIndex) -> bool:
        data = index.data(Qt.ItemDataRole.UserRole)
        return (
            index.column() == _ServiceConfigModel.VALUE_COL
            and isinstance(data, dict)
            and "widget_type" in data
            and data["widget_type"] is not None
        )



from pathlib import Path

from qgis.core import QgsSettingsTree
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QAction, QApplication

from .core.settings import Settings
from .gui.about_dialog import AboutDialog
from .gui.main_dialog import MainDialog
from .utils.plugin_utils import PluginUtils, logger


class OqtopusPlugin:

    def __init__(
        self, iface, modules_config_path=None, about_dialog_cls=None, settings_plugin_name=None
    ):
        """Constructor.

        :param iface: An interface instance that will be passed to this class
            which provides the hook by which you can manipulate the QGIS
            application at run time.
        :type iface: QgsInterface
        :param modules_config_path: Optional path to the modules configuration YAML file.
            Defaults to ``default_config.yaml`` next to this file.
        :type modules_config_path: Path | None
        :param about_dialog_cls: Optional custom About dialog class.
            Defaults to the bundled ``AboutDialog``.
        :type about_dialog_cls: type | None
        :param settings_plugin_name: Optional name for the QGIS settings tree node.
            Defaults to ``"oqtopus"``.
        :type settings_plugin_name: str | None
        """
        # Save reference to the QGIS interface
        self.iface = iface
        self.canvas = iface.mapCanvas()
        self._modules_config_path = (
            modules_config_path or Path(__file__).parent / "default_config.yaml"
        )
        self._about_dialog_cls = about_dialog_cls or AboutDialog
        self._settings_plugin_name = settings_plugin_name

        self.__version__ = PluginUtils.get_plugin_version()

        PluginUtils.init_logger()

        logger.info("")
        logger.info(f"Starting {PluginUtils.PLUGIN_NAME} plugin version {self.__version__}")

        self.actions = []
        self.main_menu_name = f"&{PluginUtils.PLUGIN_NAME}"

    # noinspection PyMethodMayBeStatic
    def tr(self, source_text):
        """
        This does not inherit from QObject but for the translation to work
        :rtype : unicode
        :param source_text: The text to translate
        :return: The translated text
        """
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QApplication.translate("OqtopusPlugin", source_text)

    def add_action(
        self,
        icon_path,
        text,
        callback,
        enabled_flag=True,
        add_to_menu=True,
        add_to_toolbar=True,
        status_tip=None,
        whats_this=None,
        parent=None,
    ):
        """Add a toolbar icon to the toolbar.

        :param icon_path: Path to the icon for this action. Can be a resource
            path (e.g. ':/plugins/foo/bar.png') or a normal file system path.
        :type icon_path: str

        :param text: Text that should be shown in menu items for this action.
        :type text: str

        :param callback: Function to be called when the action is triggered.
        :type callback: function

        :param enabled_flag: A flag indicating if the action should be enabled
            by default. Defaults to True.
        :type enabled_flag: bool

        :param add_to_menu: Flag indicating whether the action should also
            be added to the menu. Defaults to True.
        :type add_to_menu: bool

        :param add_to_toolbar: Flag indicating whether the action should also
            be added to the toolbar. Defaults to True.
        :type add_to_toolbar: bool

        :param status_tip: Optional text to show in a popup when mouse pointer
            hovers over the action.
        :type status_tip: str

        :param parent: Parent widget for the new action. Defaults None.
        :type parent: QWidget

        :param whats_this: Optional text to show in the status bar when the
            mouse pointer hovers over the action.

        :returns: The action that was created. Note that the action is also
            added to self.actions list.
        :rtype: QAction
        """

        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            # Adds plugin icon to Plugins toolbar
            self.iface.addToolBarIcon(action)

        if add_to_menu:
            self.iface.addPluginToMenu(self.main_menu_name, action)

        self.actions.append(action)

        return action

    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI."""
        # Register settings tree node early
        Settings(plugin_name=self._settings_plugin_name)

        self.add_action(
            icon_path=PluginUtils.get_plugin_icon_path("oqtopus-logo.png"),
            text=self.tr("Show &main dialog"),
            callback=self.show_main_dialog,
            parent=self.iface.mainWindow(),
            add_to_toolbar=True,
        )
        self.add_action(
            icon_path=None,
            text=self.tr("Show &log folder"),
            callback=self.show_logs_folder,
            parent=self.iface.mainWindow(),
            add_to_toolbar=False,
        )
        self.add_action(
            icon_path=PluginUtils.get_plugin_icon_path("help.svg"),
            text=self.tr("Help"),
            callback=PluginUtils.open_documentation,
            parent=self.iface.mainWindow(),
            add_to_toolbar=False,
        )
        self.add_action(
            icon_path=PluginUtils.get_plugin_icon_path("oqtopus-logo.png"),
            text=self.tr("&About"),
            callback=self.show_about_dialog,
            parent=self.iface.mainWindow(),
            add_to_toolbar=False,
        )

        self._get_main_menu_action().setIcon(
            PluginUtils.get_plugin_icon("oqtopus-logo.png"),
        )

    def unload(self):
        """Removes the plugin menu item and icon from QGIS GUI."""
        for action in self.actions:
            self.iface.removePluginMenu(self.main_menu_name, action)
            self.iface.removeToolBarIcon(action)

        # Unregister the plugin settings tree node
        plugin_name = getattr(Settings, "_plugin_name", PluginUtils.PLUGIN_ID)
        QgsSettingsTree.unregisterPluginTreeNode(plugin_name)
        Settings.instance = None

        # Remove pum modules from sys.modules to allow proper reloading
        # This is necessary because pum is in the libs folder and imported at module level
        import sys

        pum_modules = [mod for mod in sys.modules.keys() if mod.startswith("pum")]
        for mod in pum_modules:
            del sys.modules[mod]
            logger.debug(f"Removed {mod} from sys.modules for clean reload")

    def show_main_dialog(self):
        main_dialog = MainDialog(
            modules_config_path=self._modules_config_path,
            about_dialog_cls=self._about_dialog_cls,
            parent=self.iface.mainWindow(),
        )
        main_dialog.exec()

    def show_logs_folder(self):
        PluginUtils.open_logs_folder()

    def show_about_dialog(self):
        about_dialog = self._about_dialog_cls(self.iface.mainWindow())
        about_dialog.exec()

    def _get_main_menu_action(self):
        actions = self.iface.pluginMenu().actions()
        result_actions = [action for action in actions if action.text() == self.main_menu_name]

        # OSX does not support & in the menu title
        if not result_actions:
            result_actions = [
                action
                for action in actions
                if action.text() == self.main_menu_name.replace("&", "")
            ]

        return result_actions[0]



"""Background task for module install/upgrade/uninstall operations."""

import psycopg
from qgis.PyQt.QtCore import QThread, pyqtSignal

from ..libs.pum.feedback import Feedback
from ..libs.pum.pum_config import PumConfig
from ..libs.pum.upgrader import Upgrader
from ..utils.plugin_utils import logger


class ModuleOperationCanceled(Exception):
    """Exception raised when module operation is canceled."""


class ModuleOperationTask(QThread):
    """
    Background task for running module install/upgrade/uninstall operations.
    This allows the UI to remain responsive and show progress during long operations.
    """

    signalProgress = pyqtSignal(str, int, int)  # message, current, total
    signalFinished = pyqtSignal(bool, str)  # success, error_message

    def __init__(self, parent=None):
        super().__init__(parent)

        self.__pum_config = None
        self.__connection = None
        self.__operation = None  # 'install', 'upgrade', 'uninstall'
        self.__parameters = None
        self.__options = {}

        self.__feedback = None
        self.__canceled = False
        self.__error_message = None

    def start_install(
        self, pum_config: PumConfig, connection: psycopg.Connection, parameters: dict, **options
    ):
        """Start an install operation."""
        self.__pum_config = pum_config
        self.__connection = connection
        self.__operation = "install"
        self.__parameters = parameters
        self.__options = options
        self.__canceled = False
        self.__error_message = None
        self.start()

    def start_upgrade(
        self, pum_config: PumConfig, connection: psycopg.Connection, parameters: dict, **options
    ):
        """Start an upgrade operation."""
        self.__pum_config = pum_config
        self.__connection = connection
        self.__operation = "upgrade"
        self.__parameters = parameters
        self.__options = options
        self.__canceled = False
        self.__error_message = None
        self.start()

    def start_uninstall(
        self, pum_config: PumConfig, connection: psycopg.Connection, parameters: dict, **options
    ):
        """Start an uninstall operation."""
        self.__pum_config = pum_config
        self.__connection = connection
        self.__operation = "uninstall"
        self.__parameters = parameters
        self.__options = options
        self.__canceled = False
        self.__error_message = None
        self.start()

    def start_roles(
        self, pum_config: PumConfig, connection: psycopg.Connection, parameters: dict, **options
    ):
        """Start a create and grant roles operation."""
        self.__pum_config = pum_config
        self.__connection = connection
        self.__operation = "roles"
        self.__parameters = parameters
        self.__options = options
        self.__canceled = False
        self.__error_message = None
        self.start()

    def start_drop_app(
        self, pum_config: PumConfig, connection: psycopg.Connection, parameters: dict, **options
    ):
        """Start a drop app operation."""
        self.__pum_config = pum_config
        self.__connection = connection
        self.__operation = "drop_app"
        self.__parameters = parameters
        self.__options = options
        self.__canceled = False
        self.__error_message = None
        self.start()

    def start_recreate_app(
        self, pum_config: PumConfig, connection: psycopg.Connection, parameters: dict, **options
    ):
        """Start a recreate app operation."""
        self.__pum_config = pum_config
        self.__connection = connection
        self.__operation = "recreate_app"
        self.__parameters = parameters
        self.__options = options
        self.__canceled = False
        self.__error_message = None
        self.start()

    def cancel(self):
        """Cancel the current operation."""
        self.__canceled = True
        if self.__feedback:
            self.__feedback.cancel()

    def run(self):
        """Execute the operation in a background thread."""
        try:
            # Create feedback instance that emits signals
            self.__feedback = self._create_feedback()

            upgrader = Upgrader(config=self.__pum_config)

            if self.__operation == "install":
                self._run_install(upgrader)
            elif self.__operation == "upgrade":
                self._run_upgrade(upgrader)
            elif self.__operation == "uninstall":
                self._run_uninstall(upgrader)
            elif self.__operation == "roles":
                self._run_roles()
            elif self.__operation == "drop_app":
                self._run_drop_app(upgrader)
            elif self.__operation == "recreate_app":
                self._run_recreate_app(upgrader)
            else:
                raise Exception(f"Unknown operation: {self.__operation}")

            # Commit if successful and not canceled
            if not self.__canceled and self.__options.get("commit", True):
                logger.info("Committing changes to database...")
                self.__connection.commit()
                logger.info("Changes committed to the database.")

            logger.info(f"Operation '{self.__operation}' completed successfully")
            self.signalFinished.emit(True, "")

        except Exception as e:
            logger.critical(f"Module operation error in '{self.__operation}': {e}")
            logger.exception("Full traceback:")  # Log full stack trace
            self.__error_message = str(e)
            # Rollback on error
            try:
                logger.info("Rolling back transaction...")
                self.__connection.rollback()
                logger.info("Transaction rolled back")
            except Exception as rollback_error:
                logger.error(f"Rollback failed: {rollback_error}")
            self.signalFinished.emit(False, self.__error_message)

    def _run_install(self, upgrader: Upgrader):
        """Run install operation."""
        # Extract options that should not be passed to install()
        install_demo_data = self.__options.pop("install_demo_data", False)
        demo_data_name = self.__options.pop("demo_data_name", None)
        suffix = self.__options.pop("suffix", None)

        if suffix:
            # Specific-role flow: let the upgrader skip role creation,
            # then create roles ourselves with the suffix.
            self.__options["roles"] = False
            self.__options["grant"] = False

        upgrader.install(
            connection=self.__connection,
            parameters=self.__parameters,
            feedback=self.__feedback,
            commit=False,
            **self.__options,
        )

        if suffix:
            self._create_roles_with_options(suffix=suffix)

        # Install demo data if requested
        if install_demo_data and demo_data_name:
            upgrader.install_demo_data(
                connection=self.__connection,
                name=demo_data_name,
                parameters=self.__parameters,
            )

    def _run_upgrade(self, upgrader: Upgrader):
        """Run upgrade operation."""
        suffix = self.__options.pop("suffix", None)

        if suffix:
            self.__options["roles"] = False
            self.__options["grant"] = False

        upgrader.upgrade(
            connection=self.__connection,
            parameters=self.__parameters,
            feedback=self.__feedback,
            **self.__options,
        )

        if suffix:
            self._create_roles_with_options(suffix=suffix)

    def _run_uninstall(self, upgrader: Upgrader):
        """Run uninstall operation."""
        logger.info("Starting uninstall operation...")
        logger.debug(f"Parameters: {self.__parameters}")
        logger.debug(f"Options: {self.__options}")

        upgrader.uninstall(
            connection=self.__connection,
            parameters=self.__parameters,
            feedback=self.__feedback,
            commit=False,
        )

        logger.info("Uninstall operation completed")

    def _run_roles(self):
        """Run create and grant roles operation."""
        logger.info("Starting create and grant roles operation...")

        role_manager = self.__pum_config.role_manager()

        if not role_manager.roles:
            logger.warning("No roles defined in the configuration")
            return

        suffix = self.__options.pop("suffix", None)

        if suffix:
            self._create_roles_with_options(suffix=suffix)
        else:
            role_manager.create_roles(
                connection=self.__connection,
                grant=True,
                feedback=self.__feedback,
            )

        logger.info("Create and grant roles operation completed")

    def _create_roles_with_options(self, *, suffix: str | None = None):
        """Create roles using role_manager with specific-role options."""
        role_manager = self.__pum_config.role_manager()
        if not role_manager.roles:
            logger.warning("No roles defined in the configuration")
            return

        logger.info(f"Creating roles (suffix={suffix})")

        role_manager.create_roles(
            connection=self.__connection,
            suffix=suffix,
            grant=True,
            feedback=self.__feedback,
        )

    def _run_drop_app(self, upgrader: Upgrader):
        """Run drop app operation."""
        logger.info("Starting drop app operation...")
        logger.debug(f"Parameters: {self.__parameters}")
        logger.debug(f"Options: {self.__options}")

        upgrader.drop_app(
            connection=self.__connection,
            parameters=self.__parameters,
            feedback=self.__feedback,
            commit=False,
        )

        logger.info("Drop app operation completed")

    def _run_recreate_app(self, upgrader: Upgrader):
        """Run recreate app operation."""
        logger.info("Starting recreate app operation...")
        logger.debug(f"Parameters: {self.__parameters}")
        logger.debug(f"Options: {self.__options}")

        upgrader.recreate_app(
            connection=self.__connection,
            parameters=self.__parameters,
            feedback=self.__feedback,
            commit=False,
        )

        logger.info("Recreate app operation completed")

    def _create_feedback(self):
        """Create a Feedback instance that emits Qt signals."""

        class QtFeedback(Feedback):
            """Feedback implementation that emits Qt signals."""

            def __init__(self, task):
                super().__init__()
                self.task = task

            def report_progress(self, message: str, current: int = 0, total: int = 0):
                """Report progress via Qt signal.

                If current and total are provided (non-zero), use those.
                Otherwise, use the internal step counter.
                """
                # Use provided values if available, otherwise use internal counter
                if current > 0 or total > 0:
                    actual_current = current
                    actual_total = total
                else:
                    actual_current, actual_total = self.get_progress()

                logger.info(
                    f"[{actual_current}/{actual_total}] {message}" if actual_total > 0 else message
                )
                self.task.signalProgress.emit(message, actual_current, actual_total)

            def is_cancelled(self):
                """Check if operation is cancelled."""
                return self.task._ModuleOperationTask__canceled

        return QtFeedback(self)



from qgis.PyQt.QtWidgets import QApplication, QDialog, QMessageBox, QStyle

from ..core.settings import Settings
from ..utils.plugin_utils import PluginUtils

DIALOG_UI = PluginUtils.get_ui_class("settings_dialog.ui")


class SettingsDialog(QDialog, DIALOG_UI):
    def __init__(self, parent=None):
        QDialog.__init__(self, parent)
        self.setupUi(self)

        self.githubToken_lineEdit.setText(Settings().github_token.value())
        self.allow_multiple_modules_checkBox.setChecked(Settings().allow_multiple_modules.value())
        self.show_experimental_modules_checkBox.setChecked(
            Settings().show_experimental_modules.value()
        )

        # Load log column visibility settings
        self.log_show_datetime_checkBox.setChecked(Settings().log_show_datetime.value())
        self.log_show_level_checkBox.setChecked(Settings().log_show_level.value())
        self.log_show_module_checkBox.setChecked(Settings().log_show_module.value())

        self.helpButton.setIcon(
            QApplication.style().standardIcon(QStyle.StandardPixmap.SP_DialogHelpButton)
        )
        self.helpButton.clicked.connect(self.__show_github_token_help)

    def accept(self):
        Settings().github_token.setValue(self.githubToken_lineEdit.text())
        Settings().allow_multiple_modules.setValue(
            self.allow_multiple_modules_checkBox.isChecked()
        )
        Settings().show_experimental_modules.setValue(
            self.show_experimental_modules_checkBox.isChecked()
        )
        Settings().log_show_datetime.setValue(self.log_show_datetime_checkBox.isChecked())
        Settings().log_show_level.setValue(self.log_show_level_checkBox.isChecked())
        Settings().log_show_module.setValue(self.log_show_module_checkBox.isChecked())
        super().accept()

    def __show_github_token_help(self):
        QMessageBox.information(
            self,
            "GitHub Access Token Help",
            "<b>GitHub Access Token</b><br>"
            "oQtopus needs to download release data from GitHub to work properly. "
            "GitHub limits the number of requests that can be made without authentication. "
            "A personal access token is required to access private repositories or to increase API rate limits.<br><br>"
            "To generate a token:<br>"
            "1. Go to <a href='https://github.com/settings/tokens'>GitHub Personal Access Tokens</a>.<br>"
            "2. Click <b>Generate new token</b>.<br>"
            "3. Select the <code>repo</code> scope for most operations.<br>"
            "4. Copy and paste the generated token here.",
        )



import requests
from qgis.PyQt.QtCore import QDateTime, Qt

from .module_asset import ModuleAsset
from .settings import Settings


class ModulePackage:

    # enum for package type
    class Type:
        RELEASE = "release"
        BRANCH = "branch"
        PULL_REQUEST = "pull_request"
        FROM_ZIP = "from_zip"
        FROM_DIRECTORY = "from_directory"

    def __init__(
        self,
        module,
        organisation,
        repository,
        json_payload: dict,
        type=Type.RELEASE,
        name=None,
        branch=None,
        commit_sha=None,
    ):
        self.module = module
        self.organisation = organisation
        self.repository = repository
        self.type = type
        self.name = name
        self.branch = branch
        self.commit_sha = commit_sha
        self.created_at = None
        self.prerelease = False
        self.html_url = None

        self.asset_project = None
        self.asset_plugin = None

        self.source_package_zip = None
        self.source_package_dir = None

        if self.type == ModulePackage.Type.RELEASE:
            self.__parse_release(json_payload)
        elif self.type == ModulePackage.Type.BRANCH:
            pass
        elif self.type == ModulePackage.Type.PULL_REQUEST:
            self.__parse_pull_request(json_payload)
        elif self.type == ModulePackage.Type.FROM_ZIP:
            return
        elif self.type == ModulePackage.Type.FROM_DIRECTORY:
            return
        else:
            raise ValueError(f"Unknown type '{type}'")

        type = "heads"
        if self.type == ModulePackage.Type.RELEASE:
            type = "tags"

        self.download_url = f"https://github.com/{self.organisation}/{self.repository}/archive/refs/{type}/{self.branch}.zip"

    def display_name(self):
        if self.prerelease:
            return f"{self.name} (prerelease)"

        return self.name

    def fetch_commit_sha(self):
        """Fetch the latest commit SHA for the branch from GitHub API."""
        if self.type not in (ModulePackage.Type.BRANCH, ModulePackage.Type.PULL_REQUEST):
            return

        try:
            # For branches: use refs/heads/{branch}
            # For PRs: use refs/heads/{branch} from the head repo
            url = f"https://api.github.com/repos/{self.organisation}/{self.repository}/commits/{self.branch}"
            r = requests.get(url, headers=Settings.get_github_headers(), timeout=10)
            r.raise_for_status()
            commit_data = r.json()
            self.commit_sha = commit_data["sha"]
        except Exception as e:
            # If we can't fetch the commit SHA, we'll fall back to not caching
            from ..utils.plugin_utils import logger

            logger.warning(f"Failed to fetch commit SHA for branch '{self.branch}': {e}")
            self.commit_sha = None

    def __parse_release(self, json_payload: dict):
        if self.name is None:
            self.name = json_payload["name"]

        if self.name is None or self.name == "":
            self.name = json_payload["tag_name"]

        self.branch = self.name
        self.created_at = QDateTime.fromString(json_payload["created_at"], Qt.DateFormat.ISODate)
        self.prerelease = json_payload["prerelease"]
        self.html_url = json_payload["html_url"]

        # Use assets directly from the release payload (already included in releases API response)
        self.__parse_release_assets(json_payload.get("assets", []))

    def __parse_release_assets(self, json_assets: list):
        """Parse release assets from the already-fetched release data."""
        for json_asset in json_assets:
            asset = ModuleAsset(
                name=json_asset["name"],
                label=json_asset["label"],
                download_url=json_asset["browser_download_url"],
                size=json_asset["size"],
                type=None,
            )

            if asset.label == ModuleAsset.Type.PROJECT.value:
                asset.type = ModuleAsset.Type.PROJECT
                self.asset_project = asset
                continue

            if asset.label == ModuleAsset.Type.PLUGIN.value:
                asset.type = ModuleAsset.Type.PLUGIN
                self.asset_plugin = asset
                continue

            if self.asset_project and self.asset_plugin:
                # We already have all assets we need
                break

    def __parse_pull_request(self, json_payload: dict):
        if self.name is None:
            self.name = f"#{json_payload['number']} {json_payload['title']}"
        self.branch = json_payload["head"]["ref"]
        self.commit_sha = json_payload["head"]["sha"]
        self.created_at = QDateTime.fromString(json_payload["created_at"], Qt.DateFormat.ISODate)
        self.prerelease = False
        self.html_url = json_payload["html_url"]

        is_on_a_fork = json_payload["head"]["repo"]["fork"]
        if is_on_a_fork:
            self.organisation = json_payload["head"]["repo"]["owner"]["login"]
            self.repository = json_payload["head"]["repo"]["name"]



import sys
from pathlib import Path

# The qgis.PyQt shim is set up in oqtopus/__init__.py
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QApplication

from .gui.main_dialog import MainDialog
from .utils.plugin_utils import PluginUtils


def main():
    app = QApplication(sys.argv)
    icon = QIcon("oqtopus/icons/oqtopus-logo.png")
    app.setWindowIcon(icon)

    PluginUtils.init_logger()

    conf_path = Path(__file__).parent / "default_config.yaml"

    dialog = MainDialog(modules_config_path=conf_path)
    dialog.setWindowIcon(icon)
    dialog.show()

    sys.exit(app.exec())


if __name__ == "__main__":
    main()



import subprocess
import logging
from enum import Enum

from .exceptions import (
    PgDumpCommandError,
    PgDumpFailed,
    PgRestoreCommandError,
    PgRestoreFailed,
)
from .connection import format_connection_string

logger = logging.getLogger(__name__)


class DumpFormat(Enum):
    """Enumeration of supported dump formats.

    Version Added:
        1.0.0
    """

    CUSTOM = "custom"
    PLAIN = "plain"

    def to_pg_dump_flag(self):
        if self == DumpFormat.CUSTOM:
            return "-Fc"
        elif self == DumpFormat.PLAIN:
            return "-Fp"
        raise ValueError(f"Unknown dump format: {self}")


class Dumper:
    """This class is used to dump and restore a Postgres database.

    Version Added:
        1.0.0
    """

    def __init__(self, pg_connection: str, dump_path: str):
        """Initialize the Dumper.

        Args:
            pg_connection: PostgreSQL service name or connection string.
                Can be a service name (e.g., 'mydb') or a full connection string
                (e.g., 'postgresql://user:pass@host/db' or 'host=localhost dbname=mydb').
            dump_path: Path where the dump file will be saved or loaded from.

        """
        self.pg_connection = pg_connection
        self.dump_path = dump_path

    def pg_dump(
        self,
        dbname: str | None = None,
        *,
        pg_dump_exe: str = "pg_dump",
        exclude_schema: list[str] | None = None,
        format: DumpFormat = DumpFormat.CUSTOM,
    ):
        """
        Call the pg_dump command to dump a db backup

        Args:
            dbname: Name of the database to dump.
            pg_dump_exe: Path to the pg_dump executable.
            exclude_schema: List of schemas to exclude from the dump.
            format: DumpFormat, either custom (default) or plain
        """

        connection = format_connection_string(self.pg_connection)
        if dbname:
            connection = f"{connection} dbname={dbname}"

        command = [
            pg_dump_exe,
            format.to_pg_dump_flag(),
            "--no-owner",
            "--no-privileges",
            "-f",
            self.dump_path,
        ]
        if exclude_schema:
            for schema in exclude_schema:
                command.append(f"--exclude-schema={schema}")
        command.extend(["-d", connection])

        logger.debug("Running pg_dump command: %s", " ".join(command))

        try:
            output = subprocess.run(command, capture_output=True, text=True, check=False)
            if output.returncode != 0:
                logger.error("pg_dump failed: %s", output.stderr)
                raise PgDumpFailed(output.stderr)
        except TypeError:
            logger.error("Invalid command: %s", " ".join(command))
            raise PgDumpCommandError("invalid command: {}".format(" ".join(filter(None, command))))

    def pg_restore(
        self,
        dbname: str | None = None,
        pg_restore_exe: str = "pg_restore",
        exclude_schema: list[str] | None = None,
    ):
        """ """

        connection = format_connection_string(self.pg_connection)
        if dbname:
            connection = f"{connection} dbname={dbname}"

        command = [pg_restore_exe, "-d", connection, "--no-owner"]

        if exclude_schema:
            for schema in exclude_schema:
                command.append(f"--exclude-schema={schema}")
        command.append(self.dump_path)

        logger.debug("Running pg_restore command: %s", " ".join(command))

        try:
            output = subprocess.run(command, capture_output=True, text=True, check=False)
            if output.returncode != 0:
                logger.error("pg_restore failed: %s", output.stderr)
                raise PgRestoreFailed(output.stderr)
        except TypeError:
            logger.error("Invalid command: %s", " ".join(command))
            raise PgRestoreCommandError(
                "invalid command: {}".format(" ".join(filter(None, command)))
            )



import importlib
import logging
from typing import Any

# Custom SQL logging level (more verbose than DEBUG)
# Register with: logging.addLevelName(SQL, 'SQL')
SQL = 5

# Configure default logging for API usage (not CLI)
# CLI will override this with its own configuration
if not logging.getLogger().handlers:
    logging.basicConfig(
        level=logging.INFO,
        format="%(message)s",
    )


# Use relative imports so that when pum is bundled inside another package
# (e.g. oqtopus.libs.pum), the lazy imports resolve from the same package tree
# rather than picking up a different pum installed elsewhere on sys.path.
_LAZY_IMPORTS: dict[str, tuple[str, str]] = {
    "Checker": (".checker", "Checker"),
    "Changelog": (".changelog", "Changelog"),
    "configure_database_connect_access": (".database", "configure_database_connect_access"),
    "create_database": (".database", "create_database"),
    "get_database_connect_access": (".database", "get_database_connect_access"),
    "CursorResult": (".sql_content", "CursorResult"),
    "DependencyHandler": (".dependency_handler", "DependencyHandler"),
    "drop_database": (".database", "drop_database"),
    "Dumper": (".dumper", "Dumper"),
    "DumpFormat": (".dumper", "DumpFormat"),
    "HookBase": (".hook", "HookBase"),
    "Feedback": (".feedback", "Feedback"),
    "HookHandler": (".hook", "HookHandler"),
    "LogFeedback": (".feedback", "LogFeedback"),
    "ParameterDefinition": (".parameter", "ParameterDefinition"),
    "ParameterType": (".parameter", "ParameterType"),
    "Permission": (".role_manager", "Permission"),
    "PermissionType": (".role_manager", "PermissionType"),
    "PumConfig": (".pum_config", "PumConfig"),
    "Role": (".role_manager", "Role"),
    "RoleInventory": (".role_manager", "RoleInventory"),
    "RoleManager": (".role_manager", "RoleManager"),
    "RoleStatus": (".role_manager", "RoleStatus"),
    "SchemaPermissionStatus": (".role_manager", "SchemaPermissionStatus"),
    "SchemaMigrations": (".schema_migrations", "SchemaMigrations"),
    "SilentFeedback": (".feedback", "SilentFeedback"),
    "SqlContent": (".sql_content", "SqlContent"),
    "Upgrader": (".upgrader", "Upgrader"),
}

__all__ = sorted(list(_LAZY_IMPORTS.keys()) + ["SQL"])  # pyright: ignore[reportUnsupportedDunderAll]


def __getattr__(name: str) -> Any:
    if name not in _LAZY_IMPORTS:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
    module_name, symbol_name = _LAZY_IMPORTS[name]
    module = importlib.import_module(module_name, package=__package__)
    value = getattr(module, symbol_name)
    globals()[name] = value
    return value


def __dir__() -> list[str]:
    return sorted(set(list(globals().keys()) + list(_LAZY_IMPORTS.keys())))



"""Dialog for managing database CONNECT privileges."""

import logging

from qgis.PyQt.QtCore import Qt
from qgis.PyQt.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QHeaderView,
    QLabel,
    QTreeWidget,
    QTreeWidgetItem,
    QVBoxLayout,
)

from ..libs.pgserviceparser.gui.message_bar import MessageBar
from ..libs.pum.database import (
    configure_database_connect_access,
    get_database_connect_access,
)

logger = logging.getLogger(__name__)

_ROLE_NAME = Qt.ItemDataRole.UserRole


class DatabaseAccessDialog(QDialog):
    """Manage CONNECT privileges on a PostgreSQL database."""

    def __init__(self, *, connection, module_role_names: list[str], parent=None):
        super().__init__(parent)
        self._connection = connection
        self._database_name = connection.info.dbname

        self.setWindowTitle(self.tr("Configure database access"))
        self.setMinimumSize(450, 300)

        layout = QVBoxLayout(self)

        # --- Message bar ---
        self._message_bar = MessageBar(self)
        layout.addWidget(self._message_bar)

        layout.addWidget(
            QLabel(self.tr("Manage CONNECT privilege on database '%s':") % self._database_name)
        )

        # --- Query current state ---
        public_connect, roles_with_connect = get_database_connect_access(
            connection, self._database_name
        )
        roles_connect_set = set(roles_with_connect)
        module_role_set = set(module_role_names)
        other_role_names = [r for r in roles_with_connect if r not in module_role_set]

        # --- Tree ---
        self._tree = QTreeWidget(self)
        self._tree.setHeaderLabels([self.tr("Role"), self.tr("CONNECT")])
        self._tree.setRootIsDecorated(True)
        self._tree.setAlternatingRowColors(True)
        layout.addWidget(self._tree)

        self._checkable_flags = (
            Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemIsUserCheckable
            | Qt.ItemFlag.ItemIsSelectable
        )
        self._disabled_flags = Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsSelectable

        # -- PUBLIC --
        self._public_item = QTreeWidgetItem(self._tree, ["PUBLIC", ""])
        self._public_item.setFlags(self._checkable_flags)
        self._public_item.setCheckState(
            1, Qt.CheckState.Checked if public_connect else Qt.CheckState.Unchecked
        )
        self._public_item.setToolTip(
            0,
            self.tr(
                "PUBLIC is a special pseudo-role that represents all users.\n"
                "When CONNECT is granted to PUBLIC, any database user\n"
                "on this server can connect to this database.\n\n"
                "Uncheck to restrict access to specific roles only."
            ),
        )
        self._public_item.setData(0, _ROLE_NAME, "PUBLIC")

        # -- Module roles group --
        self._role_items: dict[str, QTreeWidgetItem] = {}
        if module_role_names:
            module_header = QTreeWidgetItem(self._tree, [self.tr("Module roles")])
            module_header.setFlags(Qt.ItemFlag.ItemIsEnabled)
            font = module_header.font(0)
            font.setBold(True)
            module_header.setFont(0, font)

            for name in module_role_names:
                item = QTreeWidgetItem(module_header, [name, ""])
                item.setFlags(self._checkable_flags)
                item.setCheckState(
                    1,
                    (
                        Qt.CheckState.Checked
                        if name in roles_connect_set
                        else Qt.CheckState.Unchecked
                    ),
                )
                item.setData(0, _ROLE_NAME, name)
                self._role_items[name] = item
            module_header.setExpanded(True)

        # -- Other roles group --
        if other_role_names:
            other_header = QTreeWidgetItem(self._tree, [self.tr("Other roles")])
            other_header.setFlags(Qt.ItemFlag.ItemIsEnabled)
            font = other_header.font(0)
            font.setBold(True)
            other_header.setFont(0, font)

            for name in other_role_names:
                item = QTreeWidgetItem(other_header, [name, ""])
                item.setFlags(self._checkable_flags)
                item.setCheckState(1, Qt.CheckState.Checked)
                item.setData(0, _ROLE_NAME, name)
                self._role_items[name] = item
            other_header.setExpanded(True)

        self._tree.header().setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        self._tree.header().setSectionResizeMode(1, QHeaderView.ResizeMode.ResizeToContents)

        # Disable role checkboxes when PUBLIC is checked (they're redundant)
        self._tree.itemChanged.connect(self._on_item_changed)
        self._update_role_items_enabled()

        # --- Buttons ---
        btn_box = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Apply | QDialogButtonBox.StandardButton.Close,
            self,
        )
        btn_box.button(QDialogButtonBox.StandardButton.Apply).clicked.connect(self._apply)
        btn_box.rejected.connect(self.reject)
        layout.addWidget(btn_box)

        # Baseline state for computing diffs between applies
        self._state_public = public_connect
        self._state_roles = set(roles_connect_set)

    # ------------------------------------------------------------------

    def _on_item_changed(self, item: QTreeWidgetItem, column: int):
        """React to checkbox changes in the tree."""
        if item is self._public_item and column == 1:
            self._update_role_items_enabled()

    def _update_role_items_enabled(self):
        """Disable role checkboxes when PUBLIC has CONNECT (they're redundant)."""
        public_checked = self._public_item.checkState(1) == Qt.CheckState.Checked
        flags = self._disabled_flags if public_checked else self._checkable_flags
        for item in self._role_items.values():
            item.setFlags(flags)

    def _apply(self):
        """Compute diff and apply CONNECT changes."""
        new_public = self._public_item.checkState(1) == Qt.CheckState.Checked
        do_revoke_public = self._state_public and not new_public
        do_grant_public = not self._state_public and new_public

        grant_list = [
            n
            for n, item in self._role_items.items()
            if item.checkState(1) == Qt.CheckState.Checked and n not in self._state_roles
        ]
        revoke_list = [
            n
            for n, item in self._role_items.items()
            if item.checkState(1) != Qt.CheckState.Checked and n in self._state_roles
        ]

        if not any([do_revoke_public, do_grant_public, grant_list, revoke_list]):
            self._message_bar.pushWarning(self.tr("No changes to apply."))
            return

        if do_grant_public:
            public_action = True
        elif do_revoke_public:
            public_action = False
        else:
            public_action = None

        try:
            connection_params = {"conninfo": self._connection.info.dsn}
            configure_database_connect_access(
                connection_params,
                self._database_name,
                grant_roles=grant_list or None,
                revoke_roles=revoke_list or None,
                public=public_action,
            )
            self._message_bar.pushSuccess(self.tr("Database access updated."))
            # Update baseline for next apply
            self._state_public = new_public
            self._state_roles = {
                n
                for n, item in self._role_items.items()
                if item.checkState(1) == Qt.CheckState.Checked
            }
        except Exception as exc:
            self._message_bar.pushError(
                self.tr("Failed to configure database access."), exception=exc
            )



import logging

try:
    from qgis.gui import QgsFileWidget
except ImportError:
    QgsFileWidget = None

from qgis.PyQt.QtWidgets import (
    QCheckBox,
    QComboBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QWidget,
)

from ..libs.pum import ParameterDefinition, ParameterType

logger = logging.getLogger(__name__)


class ParameterWidget(QWidget):
    def __init__(self, parameter_definition: ParameterDefinition, parent):
        QWidget.__init__(self, parent)
        self.layout = QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)
        self.value = None

        # Get the parameter type value (handle both enum and string cases)
        # This is needed because during plugin reload, enums can become strings
        param_type = parameter_definition.type
        if isinstance(param_type, ParameterType):
            param_type_value = param_type.value
        elif isinstance(param_type, str):
            # Handle string representations like "ParameterType.INTEGER" or "integer"
            param_type_value = (
                param_type.split(".")[-1].lower() if "." in param_type else param_type.lower()
            )
        else:
            param_type_value = str(param_type).split(".")[-1].lower()

        tooltip = parameter_definition.description or ""

        if param_type_value != "boolean":
            label = QLabel(parameter_definition.name, self)
            label.setToolTip(tooltip)
            self.layout.addWidget(label)

        if param_type_value == "boolean":
            self.widget = QCheckBox(parameter_definition.name, self)
            self.widget.setToolTip(tooltip)
            if parameter_definition.default is not None:
                self.widget.setChecked(parameter_definition.default)
            self.layout.addWidget(self.widget)
            self.layout.addStretch()
            self.value = lambda: self.widget.isChecked()
        elif param_type_value in ("decimal", "integer", "text"):
            if parameter_definition.values:
                self.widget = QComboBox(self)
                self.widget.setToolTip(tooltip)
                for v in parameter_definition.values:
                    self.widget.addItem(str(v), v)
                if parameter_definition.default is not None:
                    idx = self.widget.findData(parameter_definition.default)
                    if idx >= 0:
                        self.widget.setCurrentIndex(idx)
                self.layout.addWidget(self.widget)
                self.layout.addStretch()
                if param_type_value == "integer":
                    self.value = lambda: int(self.widget.currentData())
                elif param_type_value == "decimal":
                    self.value = lambda: float(self.widget.currentData())
                else:
                    self.value = lambda: self.widget.currentData()
            else:
                self.widget = QLineEdit(self)
                self.widget.setToolTip(tooltip)
                if parameter_definition.default is not None:
                    self.widget.setPlaceholderText(str(parameter_definition.default))
                self.layout.addWidget(self.widget)
                if param_type_value == "integer":
                    self.value = lambda: int(self.widget.text() or self.widget.placeholderText())
                elif param_type_value == "decimal":
                    self.value = lambda: float(self.widget.text() or self.widget.placeholderText())
                else:
                    self.value = lambda: self.widget.text() or self.widget.placeholderText()
        elif param_type_value == "path":
            if QgsFileWidget is not None:
                self.widget = QgsFileWidget(self)
                self.widget.setToolTip(tooltip)
                self.widget.setStorageMode(QgsFileWidget.StorageMode.GetFile)
                if parameter_definition.default is not None:
                    self.widget.setFilePath(str(parameter_definition.default))
                self.layout.addWidget(self.widget)
                self.value = lambda: self.widget.filePath()
            else:
                self.widget = QLineEdit(self)
                self.widget.setToolTip(tooltip)
                if parameter_definition.default is not None:
                    self.widget.setText(str(parameter_definition.default))
                self.layout.addWidget(self.widget)
                self.value = lambda: self.widget.text()
        else:
            raise ValueError(
                f"Unknown parameter type '{parameter_definition.type}' "
                f"(normalized to '{param_type_value}')"
            )



"""PostgreSQL connection service settings definitions.

See https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-PARAMKEYWORDS
"""

from enum import Enum


class SslModeEnum(Enum):
    DISABLE = "disable"
    ALLOW = "allow"
    PREFER = "prefer"
    REQUIRE = "require"
    VERIFY_CA = "verify-ca"
    VERIFY_FULL = "verify-full"


class WidgetType(Enum):
    """Widget type hints for GUI editors."""

    PLAIN = 0
    COMBOBOX = 1
    FILE = 2
    PASSWORD = 3


# Settings available for manual addition.
SERVICE_SETTINGS: dict[str, dict] = {
    "host": {
        "default": "localhost",
        "description": "Name of host to connect to.",
    },
    "port": {
        "default": "5432",
        "description": "Port number to connect to at the server host.",
    },
    "dbname": {
        "default": "test",
        "description": "The database name.",
    },
    "user": {
        "default": "",
        "description": "PostgreSQL user name to connect as.",
    },
    "password": {
        "default": "",
        "description": "Password to be used if the server demands password authentication.",
        "widget_type": WidgetType.PASSWORD,
    },
    "passfile": {
        "default": "",
        "description": "Specifies the name of the file used to store passwords.",
        "widget_type": WidgetType.FILE,
        "config": {
            "filter": "Password file (*.pgpass *.conf)",
            "title": "Select a .pgpass or .conf file",
        },
    },
    "sslmode": {
        "default": SslModeEnum.PREFER.value,
        "description": (
            "This option determines whether or with what priority a secure SSL "
            "TCP/IP connection will be negotiated with the server."
        ),
        "widget_type": WidgetType.COMBOBOX,
        "config": {"values": [e.value for e in SslModeEnum]},
    },
    "sslrootcert": {
        "default": "",
        "description": (
            "Name of a file containing SSL certificate authority (CA) certificate(s).\n"
            "If the file exists, the server's certificate will be verified to be signed "
            "by one of these authorities."
        ),
        "widget_type": WidgetType.FILE,
        "config": {
            "filter": "SSL crt files (*.crt)",
            "title": "Select the file pointing to SSL CA certificate(s)",
        },
    },
    "sslcert": {
        "default": "",
        "description": (
            "Specifies the file name of the client SSL certificate, replacing "
            "the default ~/.postgresql/postgresql.crt."
        ),
        "widget_type": WidgetType.FILE,
        "config": {
            "filter": "SSL crt files (*.crt)",
            "title": "Select the client SSL certificate file",
        },
    },
    "sslkey": {
        "default": "",
        "description": "Specifies the location for the secret key used for the client certificate.",
        "widget_type": WidgetType.FILE,
        "config": {
            "filter": "SSL secret key files (*.key)",
            "title": "Select the secret key file",
        },
    },
}

# Settings to initialize new service files.
SETTINGS_TEMPLATE: dict[str, str] = {
    "host": "localhost",
    "port": "5432",
    "dbname": "test",
}



"""Test to ensure bundled libraries are imported correctly."""

import ast
from pathlib import Path


def test_no_direct_bundled_lib_imports():
    """
    Ensure that pum and pgserviceparser are never imported directly.
    They should always be imported from oqtopus.libs to use the bundled versions.
    """
    oqtopus_dir = Path(__file__).parent.parent / "oqtopus"
    forbidden_imports = ["pum", "pgserviceparser"]
    violations = []

    libs_dir = oqtopus_dir / "libs"

    # Scan all Python files in the oqtopus directory
    for py_file in oqtopus_dir.rglob("*.py"):
        # Skip files inside the bundled lib directories (third-party code)
        if py_file.is_relative_to(libs_dir):
            continue

        try:
            with open(py_file, encoding="utf-8") as f:
                tree = ast.parse(f.read(), filename=str(py_file))

            for node in ast.walk(tree):
                # Check for "import pum" or "import pgserviceparser"
                if isinstance(node, ast.Import):
                    for alias in node.names:
                        if alias.name in forbidden_imports or any(
                            alias.name.startswith(f"{lib}.") for lib in forbidden_imports
                        ):
                            violations.append(
                                f"{py_file.relative_to(oqtopus_dir.parent)}: "
                                f"Direct import '{alias.name}' found at line {node.lineno}"
                            )

                # Check for "from pum import ..." or "from pgserviceparser import ..."
                elif isinstance(node, ast.ImportFrom):
                    if node.module in forbidden_imports or any(
                        node.module and node.module.startswith(f"{lib}.")
                        for lib in forbidden_imports
                    ):
                        violations.append(
                            f"{py_file.relative_to(oqtopus_dir.parent)}: "
                            f"Direct import 'from {node.module}' found at line {node.lineno}"
                        )

        except SyntaxError:
            # Skip files with syntax errors (might be templates or non-Python)
            continue

    if violations:
        error_msg = (
            "Found direct imports of bundled libraries. "
            "Please use 'from ..libs.pum import ...' or 'from ..libs.pgserviceparser import ...' instead:\n"
            + "\n".join(violations)
        )
        raise AssertionError(error_msg)



# standard library
import configparser
import io
import platform
import stat
from os import getenv
from pathlib import Path
from typing import Optional

# package
from .exceptions import ServiceFileNotFound, ServiceNotFound

try:
    from .gui import PGServiceParserWidget  # noqa: F401
except ImportError:
    pass


def _make_file_writable(path: Path):
    """Attempt to add write permissions to a file.

    Args:
        path: path to the file

    Raises:
        PermissionError: when the file permissions cannot be changed
    """
    current_permission = stat.S_IMODE(path.stat().st_mode)
    WRITE = stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH
    path.chmod(current_permission | WRITE)


def _when_read_only_try_to_add_write_permission(func):
    """Decorator for functions that attempt to modify the service file.

    If the file is read-only, a PermissionError exception will be raised.
    This decorator handles that error by attempting to set write permissions
    (which works if the user is the owner of the file or has proper rights to
    alter the file permissions), and rerunning the decorated function.

    If the user cannot modify permissions on the file, the PermissionError
    is re-raised.
    """

    def wrapper(*args, **kwargs):
        attempt = 0
        while attempt <= 1:
            try:
                return func(*args, **kwargs)
            except PermissionError:
                if attempt == 1:
                    raise

                try:
                    _make_file_writable(conf_path())
                except PermissionError:
                    pass
                finally:
                    attempt += 1

    return wrapper


def conf_path(create_if_missing: bool | None = False) -> Path:
    """Returns the path found for the pg_service.conf on the system as string.

    Args:
        create_if_missing: Whether to create the file (and eventually its parent folders) if the file does not exist.

    Returns:
        path to the pg_service.conf file as string
    """
    pg_config_path = None
    if getenv("PGSERVICEFILE"):
        pg_config_path = Path(getenv("PGSERVICEFILE"))
    elif getenv("PGSYSCONFDIR"):
        pg_config_path = Path(getenv("PGSYSCONFDIR")) / "pg_service.conf"
    else:
        if platform.system() == "Windows":
            pg_config_path = Path(getenv("APPDATA")) / "postgresql/.pg_service.conf"
        else:  # Linux or Darwin (Mac)
            pg_config_path = Path("~/.pg_service.conf").expanduser()

    if create_if_missing and not pg_config_path.exists():
        # Make sure all parent directories exist
        if not pg_config_path.parent.exists():
            pg_config_path.parent.mkdir(parents=True, exist_ok=True)

        # Finally, make sure the file itself exists
        Path.touch(pg_config_path)

    return pg_config_path


def full_config(conf_file_path: Path | None = None) -> configparser.ConfigParser:
    """Returns full pgservice config as configparser.ConfigParser().

    Args:
        conf_file_path: path to configuration file to load. If None the `conf_path` is used, defaults to None

    Returns:
        pg services loaded as ConfigParser

    Raises:
        ServiceFileNotFound: when the service file is not found
    """
    if conf_file_path is None:
        conf_file_path = conf_path()
    else:
        conf_file_path = Path(conf_file_path)

    if not conf_file_path.exists():
        raise ServiceFileNotFound(pg_service_filepath=conf_file_path)

    config = configparser.ConfigParser(interpolation=None)
    config.read(conf_file_path)
    return config


@_when_read_only_try_to_add_write_permission
def remove_service(service_name: str, conf_file_path: Path | None = None) -> None:
    """Remove a complete service from the service file.

    Args:
        service_name: service name
        conf_file_path: path to the pg_service.conf. If None the `conf_path()` is used,
            defaults to None

    Raises:
        ServiceFileNotFound: when the service file is not found
        ServiceNotFound: when the service is not found
        PermissionError: when the service file is read-only
    """
    config = full_config(conf_file_path)
    if service_name not in config:
        raise ServiceNotFound(
            service_name=service_name,
            existing_service_names=service_names(),
            pg_service_filepath=conf_file_path or conf_path(),
        )

    config.remove_section(service_name)
    with open(conf_file_path or conf_path(), "w") as configfile:
        config.write(configfile, space_around_delimiters=False)


@_when_read_only_try_to_add_write_permission
def rename_service(old_name: str, new_name: str, conf_file_path: Path | None = None) -> None:
    """Rename a service in the service file.

    The service settings are preserved under the new name and the old
    section is removed.

    Args:
        old_name: current service name
        new_name: desired service name
        conf_file_path: path to the pg_service.conf. If None the `conf_path()` is used,
            defaults to None

    Raises:
        ServiceFileNotFound: when the service file is not found
        ServiceNotFound: when the old service is not found
        PermissionError: when the service file is read-only
    """
    config = full_config(conf_file_path)
    if old_name not in config:
        raise ServiceNotFound(
            service_name=old_name,
            existing_service_names=service_names(),
            pg_service_filepath=conf_file_path or conf_path(),
        )

    settings = dict(config[old_name])
    config.remove_section(old_name)
    config[new_name] = settings
    with open(conf_file_path or conf_path(), "w") as configfile:
        config.write(configfile, space_around_delimiters=False)


@_when_read_only_try_to_add_write_permission
def create_service(service_name: str, settings: dict, conf_file_path: Path | None = None) -> bool:
    """Create a new service in the service file.

    If the service already exists, nothing is changed and False is returned.

    Args:
        service_name: service name
        settings: settings dict defining the service config
        conf_file_path: path to the pg_service.conf. If None the `conf_path()` is used,
            defaults to None

    Returns:
        True if the service was created, False if it already existed

    Raises:
        ServiceFileNotFound: when the service file is not found
        PermissionError: when the service file is read-only
    """
    config = full_config(conf_file_path)
    if service_name in config:
        return False

    config[service_name] = settings.copy()
    with open(conf_file_path or conf_path(), "w") as configfile:
        config.write(configfile, space_around_delimiters=False)

    return True


@_when_read_only_try_to_add_write_permission
def copy_service_settings(
    source_service_name: str,
    target_service_name: str,
    conf_file_path: Path | None = None,
) -> bool:
    """Copy all settings from one service to another.

    If the target service does not exist, it is created.

    Args:
        source_service_name: name of the service to copy settings from
        target_service_name: name of the service to copy settings to
        conf_file_path: path to the pg_service.conf. If None the `conf_path()` is used,
            defaults to None

    Returns:
        True on success

    Raises:
        ServiceFileNotFound: when the service file is not found
        ServiceNotFound: when the source service is not found
        PermissionError: when the service file is read-only
    """
    config = full_config(conf_file_path)
    if source_service_name not in config:
        raise ServiceNotFound(
            service_name=source_service_name,
            existing_service_names=service_names(),
            pg_service_filepath=conf_file_path or conf_path(),
        )

    settings = dict(config[source_service_name])
    config[target_service_name] = settings
    with open(conf_file_path or conf_path(), "w") as configfile:
        config.write(configfile, space_around_delimiters=False)

    return True


def service_config(service_name: str, conf_file_path: Path | None = None) -> dict:
    """Returns the config from the given service name as a dict.

    Args:
        service_name: service name
        conf_file_path: path to the pg_service.conf. If None the `conf_path` is used, defaults to None

    Returns:
        service settings as dictionary

    Raises:
        ServiceNotFound: when the service is not found
    """
    config = full_config(conf_file_path)

    if service_name not in config:
        raise ServiceNotFound(
            service_name=service_name,
            existing_service_names=service_names(),
            pg_service_filepath=conf_file_path or conf_path(),
        )

    return dict(config[service_name])


@_when_read_only_try_to_add_write_permission
def write_service_setting(
    service_name: str,
    setting_key: str,
    setting_value: str,
    conf_file_path: Path | None = None,
):
    """Writes a service setting to the service file.

    Args:
        service_name: service name
        setting_key: key
        setting_value: value
        conf_file_path: path to the pg_service.conf. If None the `conf_path()` is used, defaults to None

    Raises:
        ServiceFileNotFound: when the service file is not found
        ServiceNotFound: when the service is not found
        PermissionError: when the service file is read-only
    """

    config = full_config(conf_file_path)
    if service_name not in config:
        raise ServiceNotFound(
            service_name=service_name,
            existing_service_names=service_names(),
            pg_service_filepath=conf_file_path or conf_path(),
        )

    config[service_name][setting_key] = setting_value
    with open(conf_file_path or conf_path(), "w") as configfile:
        config.write(configfile, space_around_delimiters=False)


@_when_read_only_try_to_add_write_permission
def write_service(
    service_name: str, settings: dict, conf_file_path: Path | None = None, create_if_not_found: bool = False
) -> dict:
    """Writes a complete service to the service file.

    Args:
        service_name: service name
        settings: settings dict defining the service config
        conf_file_path: path to the pg_service.conf. If None the `conf_path()` is used,
            defaults to None
        create_if_not_found: option to create a new service if it does not exist yet.
            Defaults to False.

    Raises:
        ServiceFileNotFound: when the service file is not found
        ServiceNotFound: when the service is not found
        PermissionError: when the service file is read-only

    Returns:
        existing or newly created service as dictionary
    """

    config = full_config(conf_path(create_if_missing=create_if_not_found))

    if service_name not in config and not create_if_not_found:
        raise ServiceNotFound(
            service_name=service_name,
            existing_service_names=service_names(),
            pg_service_filepath=conf_file_path or conf_path(),
        )

    config[service_name] = settings.copy()
    with open(conf_file_path or conf_path(), "w") as configfile:
        config.write(configfile, space_around_delimiters=False)

    return dict(config[service_name])


def write_service_to_text(service_name: str, settings: dict) -> str:
    """Returns the complete service settings as a string.

    Args:
        service_name: service name
        settings: settings dict defining the service config

    Returns:
        Service settings as a string
    """
    config = configparser.ConfigParser(interpolation=None)
    config[service_name] = settings.copy()

    config_stream = io.StringIO()
    config.write(config_stream, space_around_delimiters=False)
    res = config_stream.getvalue()
    config_stream.close()

    return res.strip()


def service_names(conf_file_path: Path | None = None, sorted_alphabetically: bool = False) -> list[str]:
    """Returns all service names in a list.

    Args:
        conf_file_path: path to the pg_service.conf. If None the `conf_path()` is used, defaults to None
        sorted_alphabetically: whether to sort the names alphabetically (case-insensitive),
            defaults to False

    Returns:
        list of every service registered

    Raises:
        ServiceFileNotFound: when the service file is not found
    """

    config = full_config(conf_file_path)
    names = config.sections()
    return sorted(names, key=str.lower) if sorted_alphabetically else names



"""Feedback system for reporting progress and handling cancellation during operations."""

import abc
import logging

logger = logging.getLogger(__name__)


class Feedback(abc.ABC):
    """Base class for feedback during install/upgrade operations.

    This class provides methods for progress reporting and cancellation handling.
    Subclasses should implement the abstract methods to provide custom feedback mechanisms.

    Version Added:
        1.3.0
    """

    def __init__(self) -> None:
        """Initialize the feedback instance."""
        self._is_cancelled = False
        self._cancellation_locked = False
        self._current_step = 0
        self._total_steps = 0

    def set_total_steps(self, total: int) -> None:
        """Set the total number of steps for the operation.

        Args:
            total: The total number of steps.
        """
        self._total_steps = total
        self._current_step = 0

    def increment_step(self) -> None:
        """Increment the current step counter."""
        self._current_step += 1

    def get_progress(self) -> tuple[int, int]:
        """Get the current progress.

        Returns:
            A tuple of (current_step, total_steps).
        """
        return (self._current_step, self._total_steps)

    @abc.abstractmethod
    def report_progress(self, message: str, current: int = 0, total: int = 0) -> None:
        """Report progress during an operation.

        Args:
            message: A message describing the current operation.
            current: The current progress value (e.g., changelog number).
            total: The total number of steps.
        """
        pass

    def is_cancelled(self) -> bool:
        """Check if the operation has been cancelled.

        Returns:
            True if the operation should be cancelled, False otherwise.
            Always returns False if cancellation has been locked (after commit).
        """
        if self._cancellation_locked:
            return False
        return self._is_cancelled

    def cancel(self) -> None:
        """Cancel the operation.

        Note: This will have no effect if cancellation has been locked (after commit).
        """
        if not self._cancellation_locked:
            self._is_cancelled = True

    def reset(self) -> None:
        """Reset the cancellation status."""
        self._is_cancelled = False

    def lock_cancellation(self) -> None:
        """Lock cancellation to prevent it after a commit.

        Once locked, is_cancelled() will always return False and cancel() will have no effect.
        This should be called immediately before committing database changes.
        """
        self._cancellation_locked = True


class LogFeedback(Feedback):
    """Feedback implementation that logs progress messages.

    This is the default feedback implementation that simply logs messages
    without any UI interaction.

    Version Added:
        1.3.0
    """

    def report_progress(self, message: str, current: int = 0, total: int = 0) -> None:
        """Report progress by logging the message.

        Args:
            message: A message describing the current operation.
            current: The current progress value (ignored, uses internal counter).
            total: The total number of steps (ignored, uses internal counter).
        """
        logger.info(message)


class SilentFeedback(Feedback):
    """Feedback implementation that does nothing.

    This can be used when no feedback is desired.

    Version Added:
        1.3.0
    """

    def report_progress(self, message: str, current: int = 0, total: int = 0) -> None:
        """Do nothing - silent feedback.

        Args:
            message: A message describing the current operation (ignored).
            current: The current progress value (ignored).
            total: The total number of steps (ignored).
        """
        pass



"""
/***************************************************************************
 Plugin Utils
                              -------------------
        begin                : 28.4.2018
        copyright            : (C) 2018 by OPENGIS.ch
        email                : matthias@opengis.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import glob
import logging
import os
from datetime import datetime
from logging import LogRecord

from qgis.PyQt.QtCore import (
    QDir,
    QFileInfo,
    QObject,
    QSettings,
    QStandardPaths,
    QUrl,
    pyqtSignal,
)
from qgis.PyQt.QtGui import QColor, QDesktopServices, QIcon
from qgis.PyQt.uic import loadUiType

from ..libs.pum import SQL

logger = logging.getLogger("oqtopus")


class PluginUtils:

    PLUGIN_NAME = "oQtopus"
    PLUGIN_ID = "oqtopus"

    logsDirectory = ""

    COLOR_GREEN = QColor(12, 167, 137)
    COLOR_WARNING = QColor(255, 165, 0)

    DOCUMENTATION_URL = "https://opengisch.github.io/oqtopus/"

    @staticmethod
    def plugin_root_path():
        """
        Returns the root path of the plugin
        """
        return os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir))

    @staticmethod
    def plugin_cache_path():
        """Returns the persistent cache path for the plugin (GitHub API cache, downloaded packages).

        Uses CacheLocation which is appropriate for data that should persist
        across sessions but can be safely deleted.
        On macOS: ~/Library/Caches/oqtopus
        On Linux: ~/.cache/oqtopus
        On Windows: C:/Users/<USER>/AppData/Local/oqtopus/cache
        """
        cache_dir = os.path.join(
            QStandardPaths.writableLocation(QStandardPaths.StandardLocation.CacheLocation),
            "oqtopus",
        )
        os.makedirs(cache_dir, exist_ok=True)
        return cache_dir

    @staticmethod
    def plugin_temp_path():
        """Returns the temporary path for the plugin.

        DEPRECATED: Use plugin_cache_path() instead for persistent cache.
        This is kept for backward compatibility.
        """
        plugin_basename = PluginUtils.plugin_root_path().split(os.sep)[-1]

        plugin_temp_dir = os.path.join(
            QStandardPaths.writableLocation(QStandardPaths.StandardLocation.TempLocation),
            plugin_basename,
        )
        if not os.path.exists(plugin_temp_dir):
            os.makedirs(plugin_temp_dir)

        return plugin_temp_dir

    @staticmethod
    def get_all_cache_paths():
        """Returns all cache directories used by the plugin.

        This is used by the cleanup function to ensure all caches are deleted.
        """
        paths = []

        # Main cache directory (CacheLocation)
        cache_path = os.path.join(
            QStandardPaths.writableLocation(QStandardPaths.StandardLocation.CacheLocation),
            "oqtopus",
        )
        if os.path.exists(cache_path):
            paths.append(cache_path)

        # Legacy temp directory (TempLocation) - for backward compatibility
        plugin_basename = PluginUtils.plugin_root_path().split(os.sep)[-1]
        temp_path = os.path.join(
            QStandardPaths.writableLocation(QStandardPaths.StandardLocation.TempLocation),
            plugin_basename,
        )
        if os.path.exists(temp_path):
            paths.append(temp_path)

        return paths

    @staticmethod
    def get_plugin_icon_path(icon_filename):
        return os.path.join(PluginUtils.plugin_root_path(), "icons", icon_filename)

    @staticmethod
    def get_plugin_icon(icon_filename):
        return QIcon(PluginUtils.get_plugin_icon_path(icon_filename=icon_filename))

    @staticmethod
    def get_ui_class(ui_file):
        """Get UI Python class from .ui file.
           Can be filename.ui or subdirectory/filename.ui
        :param ui_file: The file of the ui in svir.ui
        :type ui_file: str
        """
        os.path.sep.join(ui_file.split("/"))
        ui_file_path = os.path.abspath(os.path.join(PluginUtils.plugin_root_path(), "ui", ui_file))
        return loadUiType(ui_file_path)[0]

    @staticmethod
    def get_metadata_file_path():
        return os.path.join(PluginUtils.plugin_root_path(), "metadata.txt")

    @staticmethod
    def get_plugin_version():
        ini_text = QSettings(PluginUtils.get_metadata_file_path(), QSettings.Format.IniFormat)
        return ini_text.value("version")

    @staticmethod
    def init_logger(logs_directory=None):
        if logs_directory is not None:
            PluginUtils.logsDirectory = logs_directory
        else:
            PluginUtils.logsDirectory = f"{PluginUtils.plugin_root_path()}/logs"

        directory = QDir(PluginUtils.logsDirectory)
        if not directory.exists():
            directory.mkpath(PluginUtils.logsDirectory)

        if directory.exists():
            # Create a new log file for each session with timestamp
            timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            logfile = QFileInfo(directory, f"oQtopus_{timestamp}.log")
            PluginUtils.currentLogFile = logfile.filePath()

            # Clean up old log files, keep only the 10 most recent
            PluginUtils._cleanup_old_logs(PluginUtils.logsDirectory, keep=10)

            # Create file handler for this session
            fileHandler = logging.FileHandler(logfile.filePath(), mode="w")
            fileHandler.setLevel(SQL)  # Use SQL level (5) to capture all messages including SQL
            fileHandler.setFormatter(logging.Formatter("%(asctime)s %(levelname)-7s %(message)s"))

            # Configure logging - basicConfig might not work if already called
            # so we configure the root logger directly
            root_logger = logging.getLogger()
            root_logger.setLevel(SQL)  # Set root to SQL level to allow SQL messages through
            root_logger.addHandler(fileHandler)

            # Set the pum library to SQL level (5) to capture all SQL statements
            # SQL (5) < DEBUG (10) < INFO (20)
            logging.getLogger("pum").setLevel(SQL)
        else:
            logger.error(f"Can't create log files directory '{PluginUtils.logsDirectory}'.")

    @staticmethod
    def _cleanup_old_logs(logs_dir: str, keep: int = 10):
        """Remove old log files, keeping only the most recent ones.

        Args:
            logs_dir: Directory containing log files
            keep: Number of most recent log files to keep
        """
        log_files = glob.glob(os.path.join(logs_dir, "oQtopus_*.log"))
        # Sort by modification time, newest first
        log_files.sort(key=os.path.getmtime, reverse=True)
        # Remove old files beyond the keep limit
        for old_file in log_files[keep:]:
            try:
                os.remove(old_file)
            except OSError:
                pass  # Ignore errors when removing old logs

    @staticmethod
    def open_logs_folder():
        QDesktopServices.openUrl(QUrl.fromLocalFile(PluginUtils.logsDirectory))

    @staticmethod
    def open_log_file():
        log_file_path = getattr(PluginUtils, "currentLogFile", None)
        if log_file_path and os.path.exists(log_file_path):
            QDesktopServices.openUrl(QUrl.fromLocalFile(log_file_path))
        else:
            # Fallback: open most recent log file
            log_files = glob.glob(os.path.join(PluginUtils.logsDirectory, "oQtopus_*.log"))
            if log_files:
                log_files.sort(key=os.path.getmtime, reverse=True)
                QDesktopServices.openUrl(QUrl.fromLocalFile(log_files[0]))
            else:
                logger.error("No log file found.")

    @staticmethod
    def open_documentation():
        QDesktopServices.openUrl(QUrl(PluginUtils.DOCUMENTATION_URL))


class LoggingBridge(logging.Handler, QObject):

    loggedLine = pyqtSignal(LogRecord, str)

    def __init__(self, level=logging.NOTSET, excluded_modules=[]):
        QObject.__init__(self)
        logging.Handler.__init__(self, level)

        self.formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(message)s")

        self.excluded_modules = excluded_modules

    def filter(self, record):
        return record.name not in self.excluded_modules

    def emit(self, record):
        log_entry = self.format(record)
        self.loggedLine.emit(record, log_entry)



"""pgserviceparser GUI - A graphical interface for managing PostgreSQL connection service files."""

from .compat import QtWidgets  # noqa: F401
from .message_bar import MessageBar, MessageLevel  # noqa: F401
from .service_widget import PGServiceParserWidget  # noqa: F401

__all__ = ["PGServiceParserWidget", "MessageBar", "MessageLevel", "QtWidgets"]



import enum
from dataclasses import dataclass, field
from typing import Optional, TYPE_CHECKING
import copy
import psycopg
import logging

from .sql_content import SqlContent
from .exceptions import PumException

if TYPE_CHECKING:
    from .feedback import Feedback

logger = logging.getLogger(__name__)


class PermissionType(enum.Enum):
    """Enum for permission types.

    Attributes:
        READ (str): Read permission.
        WRITE (str): Write permission.

    Version Added:
        1.3.0
    """

    READ = "read"
    WRITE = "write"


class Permission:
    """Class to represent a permission for a database role.

    Attributes:
        type: Type of permission (read or write).
        schemas: List of schemas this permission applies to.

    Version Added:
        1.3.0
    """

    def __init__(self, type: PermissionType | str, schemas: list[str] = None) -> None:
        if not isinstance(type, PermissionType):
            type = PermissionType(type)
        self.type = type
        self.schemas = schemas

    def grant(
        self,
        role: str,
        connection: psycopg.Connection,
        commit: bool = False,
        feedback: Optional["Feedback"] = None,
    ) -> None:
        """Grant the permission to the specified role.
        Args:
            role: The name of the role to grant the permission to.
            connection: The database connection to execute the SQL statements.
            commit: Whether to commit the transaction. Defaults to False.
            feedback: Optional feedback object for progress reporting.
        """
        if not isinstance(role, str):
            raise TypeError("Role must be a string.")

        if not self.schemas:
            raise ValueError("Schemas must be defined for the permission.")

        for schema in self.schemas:
            if feedback and feedback.is_cancelled():
                raise PumException("Permission grant cancelled by user")

            # Detect if schema exists; if not, warn and continue
            cursor = SqlContent("SELECT 1 FROM pg_namespace WHERE nspname = {schema}").execute(
                connection=connection,
                commit=False,
                parameters={"schema": psycopg.sql.Literal(schema)},
            )
            if not cursor._pum_results or not cursor._pum_results[0]:
                logger.warning(
                    f"Schema {schema} does not exist; skipping grant of {self.type.value} "
                    f"permission to role {role}."
                )
                continue

            logger.debug(
                f"Granting {self.type.value} permission on schema {schema} to role {role}."
            )
            if self.type == PermissionType.READ:
                SqlContent("""
                        GRANT USAGE ON SCHEMA {schema} TO {role};
                        GRANT SELECT, REFERENCES, TRIGGER ON ALL TABLES IN SCHEMA {schema} TO {role};
                        GRANT USAGE, SELECT ON ALL SEQUENCES IN SCHEMA {schema} TO {role};
                        GRANT EXECUTE ON ALL FUNCTIONS IN SCHEMA {schema} TO {role};
                        GRANT EXECUTE ON ALL ROUTINES IN SCHEMA {schema} TO {role};
                        ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} GRANT SELECT, REFERENCES, TRIGGER ON TABLES TO {role};
                        ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} GRANT SELECT ON SEQUENCES TO {role};
                        ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} GRANT EXECUTE ON FUNCTIONS TO {role};
                        ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} GRANT EXECUTE ON ROUTINES TO {role};
                        ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} GRANT USAGE ON TYPES TO {role};
                           """).execute(
                    connection=connection,
                    commit=False,
                    parameters={
                        "schema": psycopg.sql.Identifier(schema),
                        "role": psycopg.sql.Identifier(role),
                    },
                )
                # Grant permissions on existing types
                self._grant_existing_types(connection, schema, role, "USAGE")
            elif self.type == PermissionType.WRITE:
                SqlContent("""
                        GRANT ALL ON SCHEMA {schema} TO {role};
                        GRANT ALL ON ALL TABLES IN SCHEMA {schema} TO {role};
                        GRANT ALL ON ALL SEQUENCES IN SCHEMA {schema} TO {role};
                        GRANT ALL ON ALL FUNCTIONS IN SCHEMA {schema} TO {role};
                        GRANT ALL ON ALL ROUTINES IN SCHEMA {schema} TO {role};
                        ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} GRANT ALL ON TABLES TO {role};
                        ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} GRANT ALL ON SEQUENCES TO {role};
                        ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} GRANT ALL ON FUNCTIONS TO {role};
                        ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} GRANT ALL ON ROUTINES TO {role};
                        ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} GRANT ALL ON TYPES TO {role};
                           """).execute(
                    connection=connection,
                    commit=False,
                    parameters={
                        "schema": psycopg.sql.Identifier(schema),
                        "role": psycopg.sql.Identifier(role),
                    },
                )
                # Grant permissions on existing types
                self._grant_existing_types(connection, schema, role, "ALL")
            else:
                raise ValueError(f"Unknown permission type: {self.type}")

        if commit:
            if feedback:
                feedback.lock_cancellation()
            connection.commit()

    def _grant_existing_types(
        self, connection: psycopg.Connection, schema: str, role: str, privilege: str
    ) -> None:
        """Grant permissions on all existing types in a schema.

        Args:
            connection: The database connection.
            schema: The schema name.
            role: The role name.
            privilege: The privilege to grant (USAGE or ALL).

        """
        # Query for all types in the schema (excluding array types and internal types)
        cursor = connection.cursor()
        cursor.execute(
            """
            SELECT t.typname
            FROM pg_type t
            JOIN pg_namespace n ON t.typnamespace = n.oid
            WHERE n.nspname = %s
              AND t.typtype IN ('e', 'c', 'd', 'b', 'r')  -- enum, composite, domain, base, range
              AND t.typname NOT LIKE '_%%'  -- exclude array types
            """,
            (schema,),
        )
        types = cursor.fetchall()

        # Grant permissions on each type
        for (type_name,) in types:
            grant_sql = psycopg.sql.SQL(
                "GRANT {privilege} ON TYPE {schema}.{type_name} TO {role}"
            ).format(
                privilege=psycopg.sql.SQL(privilege),
                schema=psycopg.sql.Identifier(schema),
                type_name=psycopg.sql.Identifier(type_name),
                role=psycopg.sql.Identifier(role),
            )
            cursor.execute(grant_sql)

    def __repr__(self) -> str:
        """Return a string representation of the Permission object."""
        return f"<Permission: {self.type.value} on {self.schemas}>"


class Role:
    """
    Represents a database role with associated permissions and optional inheritance.
    The Role class encapsulates the concept of a database role, including its name,
    permissions, optional inheritance from another role, and an optional description.

    Version Added:
        1.3.0
    """

    def __init__(
        self,
        name: str,
        permissions: list[Permission] | list[str],
        *,
        inherit: Optional["Role"] = None,
        description: str | None = None,
    ) -> None:
        """Initialize the Role class.
        Args:
            name: Name of the role.
            permissions: List of permissions associated with the role.
            inherit: Optional role to inherit permissions from.
            description: Optional description of the role.
        """
        self.name = name
        if isinstance(permissions, list) and all(isinstance(p, dict) for p in permissions):
            self._permissions = [Permission(**p) for p in permissions]
        elif isinstance(permissions, list) and all(isinstance(p, Permission) for p in permissions):
            self._permissions = permissions
        else:
            raise TypeError("Permissions must be a list of dictionnaries or Permission instances.")

        if inherit is not None and not isinstance(inherit, Role):
            raise TypeError("Inherit must be a Role instance or None.")
        self.inherit = inherit
        self.description = description

    def permissions(self) -> list[Permission]:
        """
        Returns the list of permissions associated with the role.
        """
        return self._permissions

    def all_permissions(self) -> list[Permission]:
        """Return permissions from this role and all ancestors in the inheritance chain."""
        perms = list(self._permissions)
        ancestor = self.inherit
        while ancestor is not None:
            perms.extend(ancestor.permissions())
            ancestor = ancestor.inherit
        return perms

    def exists(self, connection: psycopg.Connection) -> bool:
        """Check if the role exists in the database.
        Args:
            connection: The database connection to execute the SQL statements.
        Returns:
            bool: True if the role exists, False otherwise.
        """
        cursor = SqlContent("SELECT 1 FROM pg_roles WHERE rolname = {name}").execute(
            connection=connection,
            commit=False,
            parameters={"name": psycopg.sql.Literal(self.name)},
        )
        return bool(cursor._pum_results)

    def create(
        self,
        connection: psycopg.Connection,
        grant: bool = False,
        commit: bool = False,
        feedback: Optional["Feedback"] = None,
    ) -> None:
        """Create the role in the database.
        Args:
            connection: The database connection to execute the SQL statements.
            grant: Whether to grant permissions to the role. Defaults to False.
            commit: Whether to commit the transaction. Defaults to False.
            feedback: Optional feedback object for progress reporting.
        """
        if feedback and feedback.is_cancelled():
            from .exceptions import PumException

            raise PumException("Role creation cancelled by user")

        if self.exists(connection):
            logger.debug(f"Role {self.name} already exists, skipping creation.")
        else:
            logger.debug(f"Creating role {self.name}.")
            SqlContent(
                "CREATE ROLE {name} NOSUPERUSER INHERIT NOCREATEDB NOCREATEROLE NOREPLICATION"
            ).execute(
                connection=connection,
                commit=False,
                parameters={"name": psycopg.sql.Identifier(self.name)},
            )
            if self.description:
                SqlContent("COMMENT ON ROLE {name} IS {description}").execute(
                    connection=connection,
                    commit=False,
                    parameters={
                        "name": psycopg.sql.Identifier(self.name),
                        "description": psycopg.sql.Literal(self.description),
                    },
                )
            if self.inherit:
                SqlContent("GRANT {inherit} TO {role}").execute(
                    connection=connection,
                    commit=False,
                    parameters={
                        "inherit": psycopg.sql.Identifier(self.inherit.name),
                        "role": psycopg.sql.Identifier(self.name),
                    },
                )
        if grant:
            for permission in self.permissions():
                permission.grant(
                    role=self.name, connection=connection, commit=False, feedback=feedback
                )

        if commit:
            if feedback:
                feedback.lock_cancellation()
            connection.commit()


class RoleManager:
    """
    RoleManager manages a collection of Role objects,
    allowing creation and permission management
    for multiple roles in the PostgreSQL database.

    Version Added:
        1.3.0
    """

    def __init__(self, roles=list[Role] | list[dict]) -> None:
        """Initialize the RoleManager class.:
        Args:
            roles: List of roles or dictionaries defining roles.
            Each role can be a dictionary with keys 'name', 'permissions', and optional 'description' and 'inherit'.
        """
        if isinstance(roles, list) and all(isinstance(role, dict) for role in roles):
            self.roles = {}
            for role in roles:
                _inherit = role.get("inherit")
                if _inherit is not None:
                    if _inherit not in self.roles:
                        raise ValueError(
                            f"Inherited role {_inherit} does not exist in the already defined roles. Pay attention to the order of the roles in the list."
                        )
                    role["inherit"] = self.roles[_inherit]
                self.roles[role["name"]] = Role(**role)
        elif isinstance(roles, list) and all(isinstance(role, Role) for role in roles):
            _roles = copy.deepcopy(roles)
            self.roles = {role.name: role for role in _roles}
        else:
            raise TypeError("Roles must be a list of dictionaries or Role instances.")

        for role in self.roles.values():
            if role.inherit is not None and role.inherit not in self.roles.values():
                raise ValueError(
                    f"Inherited role {role.inherit.name} does not exist in the defined roles."
                )

    def create_roles(
        self,
        connection: psycopg.Connection,
        *,
        suffix: str | None = None,
        grant: bool = False,
        commit: bool = False,
        feedback: Optional["Feedback"] = None,
    ) -> None:
        """Create roles in the database.

        When *suffix* is provided, DB-specific roles are created by appending
        the suffix to each configured role name (e.g. ``tww_user_lausanne``
        for suffix ``lausanne``). The generic (base) roles are also created,
        but no membership is granted between generic and DB-specific roles.
        This keeps DB-specific permissions isolated.

        When *suffix* is ``None`` (default), only the generic roles defined in
        the configuration are created.

        Args:
            connection: The database connection to execute the SQL statements.
            suffix: Optional suffix to append to role names for DB-specific
                roles. When provided, both the suffixed and generic roles are
                created, without cross-membership between them.
            grant: Whether to grant permissions to the roles. Defaults to False.
            commit: Whether to commit the transaction. Defaults to False.
            feedback: Optional feedback object for progress reporting.

        Version Changed:
            1.5.0: Added *suffix* parameter for DB-specific roles.
        """
        roles_list = list(self.roles.values())

        if suffix:
            for role in roles_list:
                if feedback and feedback.is_cancelled():
                    raise PumException("Role creation cancelled by user")

                specific_name = f"{role.name}_{suffix}"

                # Build a specific role with the suffixed name and same permissions
                specific_role = Role(
                    name=specific_name,
                    permissions=[
                        Permission(type=p.type, schemas=p.schemas) for p in role.permissions()
                    ],
                    description=(
                        f"{role.description} (specific to {suffix})" if role.description else None
                    ),
                )

                if feedback:
                    feedback.increment_step()
                    feedback.report_progress(f"Creating specific role: {specific_name}")

                specific_role.create(
                    connection=connection, commit=False, grant=grant, feedback=feedback
                )

                if feedback:
                    feedback.increment_step()
                    feedback.report_progress(f"Creating generic role: {role.name}")
                role.create(connection=connection, commit=False, grant=False, feedback=feedback)
        else:
            for role in roles_list:
                if feedback and feedback.is_cancelled():
                    raise PumException("Role creation cancelled by user")
                if feedback:
                    feedback.increment_step()
                    feedback.report_progress(f"Creating role: {role.name}")
                role.create(connection=connection, commit=False, grant=grant, feedback=feedback)

        if commit:
            if feedback:
                feedback.lock_cancellation()
            connection.commit()

    def grant_permissions(
        self,
        connection: psycopg.Connection,
        commit: bool = False,
        feedback: Optional["Feedback"] = None,
    ) -> None:
        """Grant permissions to the roles in the database.
        Args:
            connection: The database connection to execute the SQL statements.
            commit: Whether to commit the transaction. Defaults to False.
            feedback: Optional feedback object for progress reporting.
        """
        roles_list = list(self.roles.values())
        for role in roles_list:
            if feedback and feedback.is_cancelled():
                from .exceptions import PumException

                raise PumException("Permission grant cancelled by user")
            if feedback:
                feedback.increment_step()
                feedback.report_progress(f"Granting permissions to role: {role.name}")
            for permission in role.permissions():
                permission.grant(
                    role=role.name, connection=connection, commit=False, feedback=feedback
                )
        role_names = ", ".join(r.name for r in roles_list)
        logger.info(f"Permissions granted to roles: {role_names}.")
        if commit:
            if feedback:
                feedback.lock_cancellation()
            connection.commit()

    def _resolve_roles(
        self,
        roles: list[str] | None = None,
    ) -> list[Role]:
        """Return the list of Role objects to act on.

        When *roles* is ``None`` all configured roles are returned.
        Otherwise only the roles whose names appear in the list are
        returned (in configuration order).  Raises ``PumException``
        if any requested name is not in the configuration.

        Args:
            roles: Optional role names to filter on.

        Returns:
            Filtered list of ``Role`` objects.
        """
        if roles is None:
            return list(self.roles.values())
        unknown = set(roles) - set(self.roles)
        if unknown:
            raise PumException(
                f"Unknown role(s): {', '.join(sorted(unknown))}. "
                f"Configured roles: {', '.join(self.roles)}"
            )
        return [self.roles[name] for name in roles]

    def revoke_permissions(
        self,
        connection: psycopg.Connection,
        *,
        roles: list[str] | None = None,
        suffix: str | None = None,
        commit: bool = False,
        feedback: Optional["Feedback"] = None,
    ) -> None:
        """Revoke previously granted permissions from roles.

        When *suffix* is provided, permissions are revoked from the
        DB-specific (suffixed) roles only.  Otherwise they are revoked
        from the generic roles.

        When *roles* is provided only those configured roles are acted
        on; otherwise all configured roles are affected.

        Args:
            connection: The database connection to execute the SQL statements.
            roles: Optional list of configured role names to revoke.
                When ``None`` (default), all configured roles are revoked.
            suffix: Optional suffix identifying DB-specific roles.
            commit: Whether to commit the transaction. Defaults to False.
            feedback: Optional feedback object for progress reporting.

        Version Added:
            1.5.0
        """
        target_roles = self._resolve_roles(roles)
        for role in target_roles:
            if feedback and feedback.is_cancelled():
                raise PumException("Permission revoke cancelled by user")

            role_name = f"{role.name}_{suffix}" if suffix else role.name

            if feedback:
                feedback.increment_step()
                feedback.report_progress(f"Revoking permissions from role: {role_name}")

            for perm in role.all_permissions():
                for schema in perm.schemas or []:
                    # Check if schema exists before revoking
                    cursor = SqlContent(
                        "SELECT 1 FROM pg_namespace WHERE nspname = {schema}"
                    ).execute(
                        connection=connection,
                        commit=False,
                        parameters={"schema": psycopg.sql.Literal(schema)},
                    )
                    if not cursor._pum_results or not cursor._pum_results[0]:
                        logger.warning(
                            f"Schema {schema} does not exist; skipping revoke for role {role_name}."
                        )
                        continue

                    logger.debug(
                        f"Revoking {perm.type.value} permission on schema {schema} from role {role_name}."
                    )
                    if perm.type == PermissionType.READ:
                        SqlContent("""
                            ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} REVOKE SELECT, REFERENCES, TRIGGER ON TABLES FROM {role};
                            ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} REVOKE SELECT ON SEQUENCES FROM {role};
                            ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} REVOKE EXECUTE ON FUNCTIONS FROM {role};
                            ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} REVOKE EXECUTE ON ROUTINES FROM {role};
                            ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} REVOKE USAGE ON TYPES FROM {role};
                            REVOKE EXECUTE ON ALL ROUTINES IN SCHEMA {schema} FROM {role};
                            REVOKE EXECUTE ON ALL FUNCTIONS IN SCHEMA {schema} FROM {role};
                            REVOKE USAGE, SELECT ON ALL SEQUENCES IN SCHEMA {schema} FROM {role};
                            REVOKE SELECT, REFERENCES, TRIGGER ON ALL TABLES IN SCHEMA {schema} FROM {role};
                            REVOKE USAGE ON SCHEMA {schema} FROM {role};
                        """).execute(
                            connection=connection,
                            commit=False,
                            parameters={
                                "schema": psycopg.sql.Identifier(schema),
                                "role": psycopg.sql.Identifier(role_name),
                            },
                        )
                    elif perm.type == PermissionType.WRITE:
                        SqlContent("""
                            ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} REVOKE ALL ON TABLES FROM {role};
                            ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} REVOKE ALL ON SEQUENCES FROM {role};
                            ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} REVOKE ALL ON FUNCTIONS FROM {role};
                            ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} REVOKE ALL ON ROUTINES FROM {role};
                            ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} REVOKE ALL ON TYPES FROM {role};
                            REVOKE ALL ON ALL ROUTINES IN SCHEMA {schema} FROM {role};
                            REVOKE ALL ON ALL FUNCTIONS IN SCHEMA {schema} FROM {role};
                            REVOKE ALL ON ALL SEQUENCES IN SCHEMA {schema} FROM {role};
                            REVOKE ALL ON ALL TABLES IN SCHEMA {schema} FROM {role};
                            REVOKE ALL ON SCHEMA {schema} FROM {role};
                        """).execute(
                            connection=connection,
                            commit=False,
                            parameters={
                                "schema": psycopg.sql.Identifier(schema),
                                "role": psycopg.sql.Identifier(role_name),
                            },
                        )

        # Revoke role memberships so inherited privileges are removed too.
        # E.g. if generic is a member of specific (GRANT specific TO generic),
        # we need to REVOKE specific FROM generic for the revoke to be effective.
        for role in target_roles:
            role_name = f"{role.name}_{suffix}" if suffix else role.name

            for parent_role in RoleManager.memberships_of(
                connection=connection, user_name=role_name
            ):
                logger.debug(f"Revoking membership: REVOKE {parent_role} FROM {role_name}")
                SqlContent("REVOKE {parent} FROM {member}").execute(
                    connection=connection,
                    commit=False,
                    parameters={
                        "parent": psycopg.sql.Identifier(parent_role),
                        "member": psycopg.sql.Identifier(role_name),
                    },
                )

        role_names = ", ".join(f"{r.name}_{suffix}" if suffix else r.name for r in target_roles)
        logger.info(f"Permissions revoked from roles: {role_names}.")
        if commit:
            if feedback:
                feedback.lock_cancellation()
            connection.commit()

    def drop_roles(
        self,
        connection: psycopg.Connection,
        *,
        roles: list[str] | None = None,
        suffix: str | None = None,
        force: bool = False,
        commit: bool = False,
        feedback: Optional["Feedback"] = None,
    ) -> None:
        """Drop roles from the database.

        Permissions are revoked first (via ``revoke_permissions``), then roles
        are dropped.  When *suffix* is provided only the DB-specific roles are
        dropped; the generic roles are left untouched.  Without *suffix* only
        the generic roles are dropped.

        When *roles* is provided only those configured roles are acted
        on; otherwise all configured roles are affected.

        When *force* is ``True``, ``REASSIGN OWNED BY … TO CURRENT_USER``
        and ``DROP OWNED BY …`` are executed before the ``DROP ROLE`` to
        remove any remaining dependent objects or privileges.

        Args:
            connection: The database connection to execute the SQL statements.
            roles: Optional list of configured role names to drop.
                When ``None`` (default), all configured roles are dropped.
            suffix: Optional suffix identifying DB-specific roles.
            force: When ``True``, use ``REASSIGN OWNED`` / ``DROP OWNED``
                to remove all remaining privileges before dropping.
            commit: Whether to commit the transaction. Defaults to False.
            feedback: Optional feedback object for progress reporting.

        Version Added:
            1.5.0

        Version Changed:
            1.5.0: Added *roles* parameter for per-role operations.
        """
        target_roles = self._resolve_roles(roles)

        # Revoke permissions first so the roles own nothing
        self.revoke_permissions(
            connection=connection, roles=roles, suffix=suffix, commit=False, feedback=feedback
        )

        for role in target_roles:
            if feedback and feedback.is_cancelled():
                raise PumException("Role drop cancelled by user")

            role_name = f"{role.name}_{suffix}" if suffix else role.name

            if feedback:
                feedback.increment_step()
                feedback.report_progress(f"Dropping role: {role_name}")

            if force:
                logger.debug(f"Reassigning and dropping owned objects for role {role_name}.")
                SqlContent("REASSIGN OWNED BY {name} TO CURRENT_USER").execute(
                    connection=connection,
                    commit=False,
                    parameters={"name": psycopg.sql.Identifier(role_name)},
                )
                SqlContent("DROP OWNED BY {name}").execute(
                    connection=connection,
                    commit=False,
                    parameters={"name": psycopg.sql.Identifier(role_name)},
                )

            logger.debug(f"Dropping role {role_name}.")
            SqlContent("DROP ROLE IF EXISTS {name}").execute(
                connection=connection,
                commit=False,
                parameters={"name": psycopg.sql.Identifier(role_name)},
            )

        role_names = ", ".join(f"{r.name}_{suffix}" if suffix else r.name for r in target_roles)
        logger.info(f"Roles dropped: {role_names}.")
        if commit:
            if feedback:
                feedback.lock_cancellation()
            connection.commit()

    def grant_to(
        self,
        connection: psycopg.Connection,
        *,
        to: str,
        roles: list[str] | None = None,
        suffix: str | None = None,
        commit: bool = False,
        feedback: Optional["Feedback"] = None,
    ) -> None:
        """Grant configured roles to a database user.

        Executes ``GRANT <role> TO <to>`` for each selected role, making
        *to* a member of those roles so it inherits their permissions.

        When *suffix* is provided the suffixed role names are used
        (e.g. ``tww_viewer_lausanne``).  Otherwise the generic names are
        used.

        Args:
            connection: The database connection to execute the SQL statements.
            to: The target database role that will receive membership.
            roles: Optional list of configured role names to grant.
                When ``None`` (default), all configured roles are granted.
            suffix: Optional suffix identifying DB-specific roles.
            commit: Whether to commit the transaction. Defaults to False.
            feedback: Optional feedback object for progress reporting.

        Version Added:
            1.5.0
        """
        target_roles = self._resolve_roles(roles)

        for role in target_roles:
            if feedback and feedback.is_cancelled():
                raise PumException("Grant-to cancelled by user")

            role_name = f"{role.name}_{suffix}" if suffix else role.name

            if feedback:
                feedback.increment_step()
                feedback.report_progress(f"Granting {role_name} to {to}")

            logger.debug(f"Granting role {role_name} to {to}.")
            SqlContent("GRANT {role} TO {target}").execute(
                connection=connection,
                commit=False,
                parameters={
                    "role": psycopg.sql.Identifier(role_name),
                    "target": psycopg.sql.Identifier(to),
                },
            )

        role_names = ", ".join(f"{r.name}_{suffix}" if suffix else r.name for r in target_roles)
        logger.info(f"Roles granted to {to}: {role_names}.")
        if commit:
            if feedback:
                feedback.lock_cancellation()
            connection.commit()

    def revoke_from(
        self,
        connection: psycopg.Connection,
        *,
        from_role: str,
        roles: list[str] | None = None,
        suffix: str | None = None,
        commit: bool = False,
        feedback: Optional["Feedback"] = None,
    ) -> None:
        """Revoke configured roles from a database user.

        Executes ``REVOKE <role> FROM <from_role>`` for each selected
        role, removing *from_role*'s membership.

        When *suffix* is provided the suffixed role names are used.
        Otherwise the generic names are used.

        Args:
            connection: The database connection to execute the SQL statements.
            from_role: The target database role to revoke membership from.
            roles: Optional list of configured role names to revoke.
                When ``None`` (default), all configured roles are revoked.
            suffix: Optional suffix identifying DB-specific roles.
            commit: Whether to commit the transaction. Defaults to False.
            feedback: Optional feedback object for progress reporting.

        Version Added:
            1.5.0
        """
        target_roles = self._resolve_roles(roles)

        for role in target_roles:
            if feedback and feedback.is_cancelled():
                raise PumException("Revoke-from cancelled by user")

            role_name = f"{role.name}_{suffix}" if suffix else role.name

            if feedback:
                feedback.increment_step()
                feedback.report_progress(f"Revoking {role_name} from {from_role}")

            logger.debug(f"Revoking role {role_name} from {from_role}.")
            SqlContent("REVOKE {role} FROM {target}").execute(
                connection=connection,
                commit=False,
                parameters={
                    "role": psycopg.sql.Identifier(role_name),
                    "target": psycopg.sql.Identifier(from_role),
                },
            )

        role_names = ", ".join(f"{r.name}_{suffix}" if suffix else r.name for r in target_roles)
        logger.info(f"Roles revoked from {from_role}: {role_names}.")
        if commit:
            if feedback:
                feedback.lock_cancellation()
            connection.commit()

    @staticmethod
    def create_login_role(
        connection: psycopg.Connection,
        name: str,
        *,
        password: str | None = None,
        commit: bool = False,
    ) -> None:
        """Create a PostgreSQL role with the LOGIN attribute.

        Args:
            connection: The database connection.
            name: The name of the role to create.
            password: Optional password for the role.
            commit: Whether to commit the transaction. Defaults to False.

        Version Added:
            1.5.0
        """
        if password:
            SqlContent("CREATE ROLE {role} LOGIN PASSWORD {pwd}").execute(
                connection=connection,
                commit=commit,
                parameters={
                    "role": psycopg.sql.Identifier(name),
                    "pwd": psycopg.sql.Literal(password),
                },
            )
        else:
            SqlContent("CREATE ROLE {role} LOGIN").execute(
                connection=connection,
                commit=commit,
                parameters={"role": psycopg.sql.Identifier(name)},
            )
        logger.info(f"Login role '{name}' created.")

    @staticmethod
    def drop_login_role(
        connection: psycopg.Connection,
        name: str,
        *,
        commit: bool = False,
    ) -> None:
        """Drop a PostgreSQL login role.

        Args:
            connection: The database connection.
            name: The name of the role to drop.
            commit: Whether to commit the transaction. Defaults to False.

        Version Added:
            1.5.0
        """
        SqlContent("DROP ROLE IF EXISTS {role}").execute(
            connection=connection,
            commit=commit,
            parameters={"role": psycopg.sql.Identifier(name)},
        )
        logger.info(f"Login role '{name}' dropped.")

    @staticmethod
    def login_roles(connection: psycopg.Connection) -> list[str]:
        """Return the names of all login roles that are not superusers.

        This is useful for listing roles that can be granted membership
        in module roles.  System roles (``pg_*``) are excluded.

        Args:
            connection: The database connection.

        Returns:
            Sorted list of login role names.

        Version Added:
            1.5.0
        """
        with connection.transaction():
            cursor = connection.cursor()
            cursor.execute(
                "SELECT rolname FROM pg_roles "
                "WHERE rolcanlogin AND NOT rolsuper AND rolname NOT LIKE 'pg\\_%' "
                "ORDER BY rolname"
            )
            return [row[0] for row in cursor.fetchall()]

    @staticmethod
    def members_of(connection: psycopg.Connection, role_name: str) -> list[str]:
        """Return the login role names that are members of *role_name*.

        Args:
            connection: The database connection.
            role_name: The group role whose members to look up.

        Returns:
            Sorted list of member login role names.

        Version Added:
            1.5.4
        """
        with connection.transaction():
            cursor = connection.cursor()
            cursor.execute(
                "SELECT m.rolname "
                "FROM pg_auth_members am "
                "JOIN pg_roles r ON r.oid = am.roleid "
                "JOIN pg_roles m ON m.oid = am.member "
                "WHERE r.rolname = %s AND m.rolcanlogin "
                "ORDER BY m.rolname",
                (role_name,),
            )
            return [row[0] for row in cursor.fetchall()]

    @staticmethod
    def memberships_of(connection: psycopg.Connection, user_name: str) -> list[str]:
        """Return the role names that *user_name* is a member of.

        This is the inverse of :meth:`members_of`: instead of asking
        "who belongs to this role?", it asks "which roles does this user
        belong to?".

        Args:
            connection: The database connection.
            user_name: The login role whose memberships to look up.

        Returns:
            Sorted list of role names the user is a member of.

        Version Added:
            1.5.0
        """
        with connection.transaction():
            cursor = connection.cursor()
            cursor.execute(
                "SELECT r.rolname "
                "FROM pg_auth_members am "
                "JOIN pg_roles r ON r.oid = am.roleid "
                "JOIN pg_roles m ON m.oid = am.member "
                "WHERE m.rolname = %s "
                "ORDER BY r.rolname",
                (user_name,),
            )
            return [row[0] for row in cursor.fetchall()]

    def roles_inventory(
        self,
        connection: psycopg.Connection,
        *,
        include_superusers: bool = False,
    ) -> "RoleInventory":
        """List all database roles related to the module's configured schemas.

        Returns the module's generic roles, any DB-specific (suffixed)
        variants discovered via naming convention, and any other
        database roles that have access to the configured schemas.
        For every role the method reports which schemas it can read or
        write, whether it is a superuser, and whether it can log in.

        Args:
            connection: The database connection to use.
            include_superusers: When ``True``, superusers are included in
                the results.  Defaults to ``False`` because superusers
                implicitly have access to everything.

        Returns:
            A ``RoleInventory`` containing the discovered roles.

        Version Added:
            1.5.0
        """
        configured_schemas = set()
        for role in self.roles.values():
            for perm in role.permissions():
                if perm.schemas:
                    configured_schemas.update(perm.schemas)

        with connection.transaction():
            cursor = connection.cursor()

            # Discover all roles matching each configured name or <name>_*
            role_statuses: list[RoleStatus] = []
            known_names: set[str] = set()
            for role in self.roles.values():
                # Find the generic role and any suffixed variants
                cursor.execute(
                    "SELECT rolname FROM pg_roles WHERE rolname = %s OR rolname LIKE %s ORDER BY rolname",
                    (role.name, f"{role.name}\\_%"),
                )
                found_names = [row[0] for row in cursor.fetchall()]

                has_suffixed = any(n != role.name for n in found_names)

                for name in found_names:
                    # When suffixed variants exist, the generic role is
                    # intentionally created without schema grants (role
                    # isolation).  Skip permission expectations for it.
                    check_permissions = not (has_suffixed and name == role.name)
                    role_statuses.append(
                        _build_role_status(
                            connection,
                            name,
                            role,
                            configured_schemas,
                            check_permissions=check_permissions,
                        )
                    )
                    known_names.add(name)

            # Discover unknown roles with privileges on the configured schemas
            unknown_roles = _find_unknown_roles(
                connection,
                configured_schemas,
                known_names=known_names,
                include_superusers=include_superusers,
            )

            # Discover login roles that have no access to the configured schemas
            all_known = known_names | {r.name for r in unknown_roles}
            other_login = _find_other_login_roles(connection, configured_schemas, all_known)

            return RoleInventory(
                roles=role_statuses + unknown_roles,
                expected_roles=list(self.roles.keys()),
                other_login_roles=other_login,
            )


# ---------------------------------------------------------------------------
# Result dataclasses
# ---------------------------------------------------------------------------


@dataclass
class SchemaPermissionStatus:
    """Result of checking a single schema permission for a role.

    Version Added:
        1.5.0
    """

    schema: str
    """Name of the schema."""
    expected: PermissionType | None
    """Expected permission type from the configuration, or ``None``."""
    has_read: bool = False
    """Whether the role currently has read-level access."""
    has_write: bool = False
    """Whether the role currently has write-level access."""

    @property
    def satisfied(self) -> bool:
        """``True`` when the actual privileges match the expected ones."""
        if self.expected == PermissionType.READ:
            return self.has_read
        if self.expected == PermissionType.WRITE:
            return self.has_write
        # No expectation – anything is fine
        return True


@dataclass
class RoleStatus:
    """A database role discovered during ``RoleManager.check_roles``.

    Only roles that actually exist in PostgreSQL are represented.

    Version Added:
        1.5.0
    """

    name: str
    """Role name in PostgreSQL."""
    role: Role | None = None
    """Configured role this maps to, or ``None`` for unknown roles."""
    suffix: str = ""
    """DB-specific suffix (e.g. ``"lausanne"``), empty for generic roles."""
    schema_permissions: list[SchemaPermissionStatus] = field(default_factory=list)
    """Per-schema permission details."""
    granted_to: list[str] = field(default_factory=list)
    """Roles that this role is a member of (i.e. ``GRANT parent TO this``)."""
    superuser: bool = False
    """Whether the role is a PostgreSQL superuser."""
    login: bool = False
    """Whether the role has the LOGIN attribute."""

    @property
    def is_unknown(self) -> bool:
        """``True`` when the role is not mapped to a configuration entry."""
        return self.role is None

    @property
    def is_suffixed(self) -> bool:
        """``True`` when the role is a DB-specific (suffixed) variant."""
        return self.suffix != ""

    @property
    def schemas(self) -> list[str]:
        """List of schema names from the permission details."""
        return [sp.schema for sp in self.schema_permissions]


@dataclass
class RoleInventory:
    """Result of ``RoleManager.roles_inventory``.

    Contains all discovered database roles related to the module's
    configured schemas: configured roles (generic and suffixed),
    other roles with schema access, and the list of expected role
    names from the configuration.

    Version Added:
        1.5.0
    """

    roles: list[RoleStatus] = field(default_factory=list)
    """All discovered roles."""
    expected_roles: list[str] = field(default_factory=list)
    """Role names from the configuration that were expected to exist."""
    other_login_roles: list[str] = field(default_factory=list)
    """Login roles that are not superusers and have no access to any configured schema.

    Version Added:
        1.5.0
    """

    @property
    def configured_roles(self) -> list["RoleStatus"]:
        """Roles that are mapped to a configuration entry."""
        return [r for r in self.roles if not r.is_unknown]

    @property
    def grantee_roles(self) -> list["RoleStatus"]:
        """Roles not in the configuration but that are members of a configured role.

        These are typically login users that were granted module roles
        via ``grant_to`` and inherit schema access through membership.
        """
        configured_names = {r.name for r in self.roles if not r.is_unknown}
        return [
            r
            for r in self.roles
            if r.is_unknown and any(g in configured_names for g in r.granted_to)
        ]

    @property
    def unknown_roles(self) -> list["RoleStatus"]:
        """Roles not in the configuration that have schema access and are not grantees."""
        configured_names = {r.name for r in self.roles if not r.is_unknown}
        return [
            r
            for r in self.roles
            if r.is_unknown and not any(g in configured_names for g in r.granted_to)
        ]

    @property
    def missing_roles(self) -> list[str]:
        """Configured role names for which no DB role was found."""
        found = {r.role.name for r in self.configured_roles}
        return [name for name in self.expected_roles if name not in found]


# ---------------------------------------------------------------------------
# Internal helpers
# ---------------------------------------------------------------------------


def _build_role_status(
    connection: psycopg.Connection,
    name: str,
    role: Role,
    configured_schemas: set[str],
    *,
    check_permissions: bool = True,
) -> RoleStatus:
    """Build a ``RoleStatus`` for a role known to exist in the database.

    When *check_permissions* is ``False`` the schema permission entries
    are still created but with ``expected=None`` so they never show as
    mismatched.  This is used for generic roles when suffixed variants
    exist (role isolation: the generic role is intentionally not granted
    schema permissions).
    """
    cursor = connection.cursor()

    # Fetch role attributes
    cursor.execute(
        "SELECT rolcanlogin FROM pg_roles WHERE rolname = %s",
        (name,),
    )
    row = cursor.fetchone()
    can_login = bool(row and row[0])

    # Discover role memberships (GRANT <parent> TO <name>)
    granted_to = RoleManager.memberships_of(connection=connection, user_name=name)

    # Check every configured schema
    schema_statuses: list[SchemaPermissionStatus] = []
    for perm in role.permissions():
        for schema in perm.schemas or []:
            has_read = _has_read(connection, name, schema)
            has_write = _has_write(connection, name, schema)
            schema_statuses.append(
                SchemaPermissionStatus(
                    schema=schema,
                    expected=perm.type if check_permissions else None,
                    has_read=has_read,
                    has_write=has_write,
                )
            )

    return RoleStatus(
        name=name,
        role=role,
        suffix=name[len(role.name) + 1 :] if name != role.name else "",
        schema_permissions=schema_statuses,
        granted_to=granted_to,
        login=can_login,
    )


def _has_read(connection: psycopg.Connection, role: str, schema: str) -> bool:
    """Return ``True`` if *role* has read-level access to *schema*.

    Checks USAGE on the schema plus SELECT on all tables (if any exist).
    """
    cursor = connection.cursor()
    cursor.execute("SELECT has_schema_privilege(%s, %s, 'USAGE')", (role, schema))
    row = cursor.fetchone()
    if not row or not row[0]:
        return False

    # Check SELECT on all tables/views
    cursor.execute(
        """
        SELECT c.relname
        FROM pg_class c
        JOIN pg_namespace n ON n.oid = c.relnamespace
        WHERE n.nspname = %s AND c.relkind IN ('r', 'v', 'm')
        LIMIT 1
        """,
        (schema,),
    )
    if cursor.fetchone() is not None:
        cursor.execute(
            """
            SELECT bool_and(has_table_privilege(%s, n.nspname || '.' || c.relname, 'SELECT'))
            FROM pg_class c
            JOIN pg_namespace n ON n.oid = c.relnamespace
            WHERE n.nspname = %s AND c.relkind IN ('r', 'v', 'm')
            """,
            (role, schema),
        )
        row = cursor.fetchone()
        if not row or not row[0]:
            return False

    return True


def _has_write(connection: psycopg.Connection, role: str, schema: str) -> bool:
    """Return ``True`` if *role* has write-level access to *schema*.

    Checks CREATE on the schema plus INSERT/UPDATE/DELETE on all tables
    (if any exist).
    """
    cursor = connection.cursor()
    cursor.execute("SELECT has_schema_privilege(%s, %s, 'CREATE')", (role, schema))
    row = cursor.fetchone()
    if not row or not row[0]:
        return False

    cursor.execute(
        """
        SELECT c.relname
        FROM pg_class c
        JOIN pg_namespace n ON n.oid = c.relnamespace
        WHERE n.nspname = %s AND c.relkind = 'r'
        LIMIT 1
        """,
        (schema,),
    )
    if cursor.fetchone() is not None:
        cursor.execute(
            """
            SELECT bool_and(
                has_table_privilege(%s, n.nspname || '.' || c.relname, 'INSERT')
                AND has_table_privilege(%s, n.nspname || '.' || c.relname, 'UPDATE')
                AND has_table_privilege(%s, n.nspname || '.' || c.relname, 'DELETE')
            )
            FROM pg_class c
            JOIN pg_namespace n ON n.oid = c.relnamespace
            WHERE n.nspname = %s AND c.relkind = 'r'
            """,
            (role, role, role, schema),
        )
        row = cursor.fetchone()
        if not row or not row[0]:
            return False

    return True


def _find_unknown_roles(
    connection: psycopg.Connection,
    schemas: set[str],
    known_names: set[str],
    *,
    include_superusers: bool = False,
) -> list[RoleStatus]:
    """Return roles not in *known_names* that have privileges on *schemas*.

    For each unknown role, all *schemas* are checked for read/write access
    so the caller can see which schemas it can or cannot reach.

    Excludes PostgreSQL built-in roles (``pg_*``).

    Args:
        include_superusers: When ``True``, superusers are included in the
            result.  When ``False`` (default), they are filtered out.
    """
    if not schemas:
        return []

    cursor = connection.cursor()
    if include_superusers:
        cursor.execute(
            """
            SELECT DISTINCT r.rolname, r.rolsuper, r.rolcanlogin
            FROM pg_namespace n
            CROSS JOIN pg_roles r
            WHERE n.nspname = ANY(%s)
              AND (r.rolsuper OR has_schema_privilege(r.rolname, n.nspname, 'USAGE'))
              AND r.rolname NOT LIKE 'pg_%%'
            ORDER BY r.rolname
            """,
            (list(schemas),),
        )
    else:
        cursor.execute(
            """
            SELECT DISTINCT r.rolname, r.rolsuper, r.rolcanlogin
            FROM pg_namespace n
            CROSS JOIN pg_roles r
            WHERE n.nspname = ANY(%s)
              AND has_schema_privilege(r.rolname, n.nspname, 'USAGE')
              AND NOT r.rolsuper
              AND r.rolname NOT LIKE 'pg_%%'
            ORDER BY r.rolname
            """,
            (list(schemas),),
        )

    role_info: dict[str, tuple[bool, bool]] = {}
    for rolname, is_super, can_login in cursor.fetchall():
        if rolname not in known_names:
            role_info[rolname] = (is_super, can_login)

    # For each unknown role, check all configured schemas
    results: list[RoleStatus] = []
    for name, (is_super, can_login) in role_info.items():
        schema_perms = []
        for schema in sorted(schemas):
            has_read = _has_read(connection, name, schema)
            has_write = _has_write(connection, name, schema)
            schema_perms.append(
                SchemaPermissionStatus(
                    schema=schema,
                    expected=None,
                    has_read=has_read,
                    has_write=has_write,
                )
            )

        # Discover memberships (which roles was this role granted?)
        granted_to = RoleManager.memberships_of(connection=connection, user_name=name)

        results.append(
            RoleStatus(
                name=name,
                superuser=is_super,
                login=can_login,
                schema_permissions=schema_perms,
                granted_to=granted_to,
            )
        )

    return results


def _find_other_login_roles(
    connection: psycopg.Connection,
    schemas: set[str],
    known_names: set[str],
) -> list[str]:
    """Return login roles that are not superusers and have no access to any *schema*.

    These are database roles that can authenticate but do not have USAGE
    on any of the module's configured schemas.  Built-in PostgreSQL roles
    (``pg_*``) are excluded.

    Args:
        connection: The database connection to use.
        schemas: Set of configured schema names.
        known_names: Role names already accounted for (configured + unknown
            with schema access).  These are excluded from the result.

    Returns:
        Sorted list of role names.
    """
    cursor = connection.cursor()
    cursor.execute(
        """
        SELECT r.rolname
        FROM pg_roles r
        WHERE r.rolcanlogin
          AND NOT r.rolsuper
          AND r.rolname NOT LIKE 'pg_%%'
        ORDER BY r.rolname
        """,
    )
    login_roles = [row[0] for row in cursor.fetchall()]

    results: list[str] = []
    for name in login_roles:
        if name in known_names:
            continue
        # Check that the role has no USAGE on any configured schema
        has_access = False
        for schema in schemas:
            cursor.execute(
                "SELECT has_schema_privilege(%s, %s, 'USAGE')",
                (name, schema),
            )
            if cursor.fetchone()[0]:
                has_access = True
                break
        if not has_access:
            results.append(name)

    return results



from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
import re

import psycopg

from .connection import format_connection_string


class DifferenceType(Enum):
    """Type of difference found."""

    ADDED = "added"
    REMOVED = "removed"


@dataclass
class DifferenceItem:
    """Represents a single difference between databases."""

    type: DifferenceType
    content: dict | str  # dict for structured data, str for backward compatibility

    def __str__(self) -> str:
        """String representation with marker."""
        marker = "+" if self.type == DifferenceType.ADDED else "-"
        if isinstance(self.content, dict):
            # For structured content, create a readable string
            return f"{marker} {self.content}"
        return f"{marker} {self.content}"


@dataclass
class CheckResult:
    """Result of a single check (e.g., tables, columns)."""

    name: str
    key: str
    passed: bool
    differences: list[DifferenceItem] = field(default_factory=list)

    @property
    def difference_count(self) -> int:
        """Number of differences found."""
        return len(self.differences)


@dataclass
class ComparisonReport:
    """Complete database comparison report."""

    pg_connection1: str
    pg_connection2: str
    timestamp: datetime
    check_results: list[CheckResult] = field(default_factory=list)

    @property
    def passed(self) -> bool:
        """Whether all checks passed."""
        return all(result.passed for result in self.check_results)

    @property
    def total_checks(self) -> int:
        """Total number of checks performed."""
        return len(self.check_results)

    @property
    def passed_checks(self) -> int:
        """Number of checks that passed."""
        return sum(1 for result in self.check_results if result.passed)

    @property
    def failed_checks(self) -> int:
        """Number of checks that failed."""
        return self.total_checks - self.passed_checks

    @property
    def total_differences(self) -> int:
        """Total number of differences across all checks."""
        return sum(result.difference_count for result in self.check_results)


class Checker:
    """This class is used to compare 2 Postgres databases and show the
    differences.
    """

    def __init__(
        self,
        pg_connection1,
        pg_connection2,
        exclude_schema=None,
        exclude_field_pattern=None,
        ignore_list=None,
    ):
        """Initialize the Checker.

        Args:
            pg_connection1: PostgreSQL service name or connection string for the first database.
                Can be a service name (e.g., 'mydb') or a full connection string
                (e.g., 'postgresql://user:pass@host/db' or 'host=localhost dbname=mydb').
            pg_connection2: PostgreSQL service name or connection string for the second database.
            exclude_schema: List of schemas to be ignored in check.
            exclude_field_pattern: List of field patterns to be ignored in check.
            ignore_list: List of elements to be ignored in check (ex. tables, columns,
                views, ...).
        """
        self.pg_connection1 = pg_connection1
        self.pg_connection2 = pg_connection2

        self.conn1 = psycopg.connect(format_connection_string(pg_connection1))
        self.cur1 = self.conn1.cursor()

        self.conn2 = psycopg.connect(format_connection_string(pg_connection2))
        self.cur2 = self.conn2.cursor()

        self.ignore_list = ignore_list or []
        self.exclude_schema = "('information_schema'"
        if exclude_schema is not None:
            for schema in exclude_schema:
                self.exclude_schema += f", '{schema}'"
        self.exclude_schema += ")"
        self.exclude_field_pattern = exclude_field_pattern or []

    def run_checks(self) -> ComparisonReport:
        """Run all the checks functions.

        Returns:
            Complete comparison report with all check results.
        """
        checks = [
            ("tables", "Tables", self.check_tables),
            ("columns", "Columns", lambda: self.check_columns("views" not in self.ignore_list)),
            ("constraints", "Constraints", self.check_constraints),
            ("views", "Views", self.check_views),
            ("sequences", "Sequences", self.check_sequences),
            ("indexes", "Indexes", self.check_indexes),
            ("triggers", "Triggers", self.check_triggers),
            ("functions", "Functions", self.check_functions),
            ("rules", "Rules", self.check_rules),
        ]

        check_results = []
        for check_key, check_name, check_func in checks:
            if check_key not in self.ignore_list:
                passed, differences = check_func()
                check_results.append(
                    CheckResult(
                        name=check_name,
                        key=check_key,
                        passed=passed,
                        differences=differences,
                    )
                )

        return ComparisonReport(
            pg_connection1=self.pg_connection1,
            pg_connection2=self.pg_connection2,
            timestamp=datetime.now(),
            check_results=check_results,
        )

    def check_tables(self):
        """Check if the tables are equals.

        Returns:
            tuple: A tuple containing:
                - bool: True if the tables are the same, False otherwise.
                - list: A list with the differences.
        """
        query = rf"""SELECT table_schema, table_name
                FROM information_schema.tables
                WHERE table_schema NOT IN {self.exclude_schema}
                    AND table_schema NOT LIKE 'pg\_%'
                    AND table_type NOT LIKE 'VIEW'
                ORDER BY table_schema, table_name
                """

        return self.__check_equals(query)

    def check_columns(self, check_views=True):
        """Check if the columns in all tables are equals.

        Args:
            check_views: If True, check the columns of all the tables and views,
                if False check only the columns of the tables.

        Returns:
            tuple: A tuple containing:
                - bool: True if the columns are the same, False otherwise.
                - list: A list with the differences.
        """
        # First, get the list of tables that exist in BOTH databases
        # to avoid reporting columns from tables that don't exist in one DB
        if check_views:
            table_query = rf"""SELECT table_schema, table_name
                         FROM information_schema.tables
                         WHERE table_schema NOT IN {self.exclude_schema}
                            AND table_schema NOT LIKE 'pg\_%'
                         ORDER BY table_schema,table_name
                         """
        else:
            table_query = rf"""SELECT table_schema, table_name
                         FROM information_schema.tables
                         WHERE table_schema NOT IN {self.exclude_schema}
                            AND table_schema NOT LIKE 'pg\_%'
                            AND table_type NOT LIKE 'VIEW'
                         ORDER BY table_schema,table_name
                         """

        # Get tables from both databases
        self.cur1.execute(table_query)
        tables1 = set(self.cur1.fetchall())

        self.cur2.execute(table_query)
        tables2 = set(self.cur2.fetchall())

        # Only check columns for tables that exist in both databases
        common_tables = tables1.intersection(tables2)

        if not common_tables:
            # No common tables, so no columns to compare
            return True, []

        # Build the WHERE clause to only include common tables
        table_conditions = " OR ".join(
            [
                f"(isc.table_schema = '{schema}' AND isc.table_name = '{table}')"
                for schema, table in common_tables
            ]
        )

        query = f"""
                SELECT isc.table_schema, isc.table_name, column_name,
                    column_default, is_nullable, data_type,
                    character_maximum_length::text, numeric_precision::text,
                    numeric_precision_radix::text, datetime_precision::text
                FROM information_schema.columns isc
                WHERE ({table_conditions})
                    {("".join([f" AND column_name NOT LIKE '{pattern}'" for pattern in self.exclude_field_pattern]))}
                ORDER BY isc.table_schema, isc.table_name, column_name
                """

        return self.__check_equals(query)

    def check_constraints(self):
        """Check if the constraints are equals.

        Returns:
            tuple: A tuple containing:
                - bool: True if the constraints are the same, False otherwise.
                - list: A list with the differences.
        """
        # Get tables from both databases to filter constraints
        table_query = f"""SELECT table_schema, table_name
                         FROM information_schema.tables
                         WHERE table_schema NOT IN {self.exclude_schema}
                            AND table_schema NOT LIKE 'pg\\_%'
                            AND table_type NOT LIKE 'VIEW'
                         ORDER BY table_schema,table_name
                         """

        self.cur1.execute(table_query)
        tables1 = set(self.cur1.fetchall())

        self.cur2.execute(table_query)
        tables2 = set(self.cur2.fetchall())

        # Only check constraints for tables that exist in both databases
        common_tables = tables1.intersection(tables2)

        if not common_tables:
            return True, []

        # Build the WHERE clause to only include common tables
        table_conditions = " OR ".join(
            [
                f"(tc.constraint_schema = '{schema}' AND tc.table_name = '{table}')"
                for schema, table in common_tables
            ]
        )

        # Build WHERE clause for CHECK constraints
        check_table_conditions = " OR ".join(
            [
                f"(n.nspname = '{schema}' AND cl.relname = '{table}')"
                for schema, table in common_tables
            ]
        )

        # Query for KEY constraints (PRIMARY KEY, FOREIGN KEY, UNIQUE)
        key_query = f"""
                    SELECT
                        tc.constraint_name,
                        tc.constraint_schema || '.' || tc.table_name || '.' ||
                            kcu.column_name as physical_full_name,
                        tc.constraint_schema,
                        tc.table_name,
                        kcu.column_name,
                        ccu.table_name as foreign_table_name,
                        ccu.column_name as foreign_column_name,
                        tc.constraint_type,
                        pg_get_constraintdef((
                            SELECT con.oid FROM pg_constraint con
                            JOIN pg_namespace nsp ON con.connamespace = nsp.oid
                            WHERE con.conname = tc.constraint_name
                            AND nsp.nspname = tc.constraint_schema
                            LIMIT 1
                        )) as constraint_definition
                    FROM information_schema.table_constraints as tc
                    JOIN information_schema.key_column_usage as kcu ON
                        (tc.constraint_name = kcu.constraint_name AND
                        tc.table_name = kcu.table_name)
                    JOIN information_schema.constraint_column_usage as ccu ON
                        ccu.constraint_name = tc.constraint_name
                    WHERE ({table_conditions})
                    ORDER BY tc.constraint_schema, physical_full_name,
                        tc.constraint_name, foreign_table_name,
                        foreign_column_name
                    """

        # Query for CHECK constraints (they don't appear in key_column_usage)
        check_query = f"""
                    SELECT
                        c.conname as constraint_name,
                        n.nspname || '.' || cl.relname as physical_full_name,
                        n.nspname as constraint_schema,
                        cl.relname as table_name,
                        '' as column_name,
                        '' as foreign_table_name,
                        '' as foreign_column_name,
                        'CHECK' as constraint_type,
                        pg_get_constraintdef(c.oid) as constraint_definition
                    FROM pg_constraint c
                    JOIN pg_class cl ON c.conrelid = cl.oid
                    JOIN pg_namespace n ON cl.relnamespace = n.oid
                    WHERE c.contype = 'c'
                        AND ({check_table_conditions})
                    ORDER BY n.nspname, cl.relname, c.conname
                    """

        # Normalization function for constraint records
        def normalize_constraint_record(record_dict, col_names):
            """Normalize constraint definitions in a record."""
            normalized = record_dict.copy()
            if "constraint_definition" in normalized and normalized["constraint_definition"]:
                normalized["constraint_definition"] = self.__normalize_constraint_definition(
                    normalized["constraint_definition"]
                )
            return normalized

        # Execute both queries and combine results
        passed_keys, diffs_keys = self.__check_equals(
            key_query, normalize_func=normalize_constraint_record
        )
        passed_checks, diffs_checks = self.__check_equals(
            check_query, normalize_func=normalize_constraint_record
        )

        return (passed_keys and passed_checks, diffs_keys + diffs_checks)

    def check_views(self):
        """Check if the views are equals.

        Returns:
            tuple: A tuple containing:
                - bool: True if the views are the same, False otherwise.
                - list: A list with the differences.
        """
        query = rf"""
        SELECT table_schema, table_name, REPLACE(view_definition,'"','')
        FROM INFORMATION_SCHEMA.views
        WHERE table_schema NOT IN {self.exclude_schema}
        AND table_schema NOT LIKE 'pg\_%'
        AND table_name not like 'vw_export_%'
        ORDER BY table_schema, table_name
        """

        return self.__check_equals(query)

    def check_sequences(self):
        """Check if the sequences are equals.

        Returns:
            tuple: A tuple containing:
                - bool: True if the sequences are the same, False otherwise.
                - list: A list with the differences.
        """
        query = f"""
        SELECT c.relname,
               ns.nspname as schema_name
        FROM pg_class c
        JOIN pg_namespace ns ON c.relnamespace = ns.oid
        WHERE c.relkind = 'S'
              AND ns.nspname NOT IN {self.exclude_schema}
        ORDER BY c.relname"""

        return self.__check_equals(query)

    def check_indexes(self):
        """Check if the indexes are equals.

        Returns:
            tuple: A tuple containing:
                - bool: True if the indexes are the same, False otherwise.
                - list: A list with the differences.
        """
        # Get tables from both databases to filter indexes
        table_query = f"""SELECT table_schema, table_name
                         FROM information_schema.tables
                         WHERE table_schema NOT IN {self.exclude_schema}
                            AND table_schema NOT LIKE 'pg\\_%'
                            AND table_type NOT LIKE 'VIEW'
                         ORDER BY table_schema,table_name
                         """

        self.cur1.execute(table_query)
        tables1 = set(self.cur1.fetchall())

        self.cur2.execute(table_query)
        tables2 = set(self.cur2.fetchall())

        # Only check indexes for tables that exist in both databases
        common_tables = tables1.intersection(tables2)

        if not common_tables:
            return True, []

        # Build the WHERE clause to only include common tables
        table_conditions = " OR ".join(
            [
                f"(ns.nspname = '{schema}' AND t.relname = '{table}')"
                for schema, table in common_tables
            ]
        )

        query = rf"""
        SELECT
            ns.nspname as schema_name,
            t.relname as table_name,
            i.relname as index_name,
            a.attname as column_name,
            pg_get_indexdef(i.oid) as index_definition
        FROM
            pg_class t,
            pg_class i,
            pg_index ix,
            pg_attribute a,
            pg_namespace ns
        WHERE
            t.oid = ix.indrelid
            AND i.oid = ix.indexrelid
            AND a.attrelid = t.oid
            AND t.relnamespace = ns.oid
            AND a.attnum = ANY(ix.indkey)
            AND t.relkind = 'r'
            AND ({table_conditions})
        ORDER BY
            ns.nspname,
            t.relname,
            i.relname,
            a.attname
        """
        return self.__check_equals(query)

    def check_triggers(self):
        """Check if the triggers are equals.

        Returns:
            tuple: A tuple containing:
                - bool: True if the triggers are the same, False otherwise.
                - list: A list with the differences.
        """
        query = f"""
        WITH trigger_list AS (
            select tgname, tgisinternal from pg_trigger
            GROUP BY tgname, tgisinternal
        )
        select ns.nspname as schema_name, p.relname, t.tgname, pp.prosrc
        from pg_trigger t, pg_proc pp, trigger_list tl, pg_class p, pg_namespace ns
        where pp.oid = t.tgfoid
            and t.tgname = tl.tgname
            AND t.tgrelid = p.oid
            AND p.relnamespace = ns.oid
            AND NOT tl.tgisinternal
            AND ns.nspname NOT IN {self.exclude_schema}
        ORDER BY p.relname, t.tgname, pp.prosrc"""

        return self.__check_equals(query)

    def check_functions(self):
        """Check if the functions are equals.

        Returns:
            tuple: A tuple containing:
                - bool: True if the functions are the same, False otherwise.
                - list: A list with the differences.
        """
        query = rf"""
        SELECT routines.routine_schema, routines.routine_name, parameters.data_type,
            routines.routine_definition
        FROM information_schema.routines
        LEFT JOIN information_schema.parameters
        ON routines.specific_name=parameters.specific_name
        WHERE routines.specific_schema NOT IN {self.exclude_schema}
            AND routines.specific_schema NOT LIKE 'pg\_%'
            AND routines.specific_schema <> 'information_schema'
        ORDER BY routines.routine_name, parameters.data_type,
            routines.routine_definition, parameters.ordinal_position
            """

        return self.__check_equals(query)

    def check_rules(self):
        """Check if the rules are equals.

        Returns:
            tuple: A tuple containing:
                - bool: True if the rules are the same, False otherwise.
                - list: A list with the differences.
        """
        query = rf"""
        select n.nspname as rule_schema,
        c.relname as rule_table,
        r.rulename as rule_name,
        case r.ev_type
            when '1' then 'SELECT'
            when '2' then 'UPDATE'
            when '3' then 'INSERT'
            when '4' then 'DELETE'
            else 'UNKNOWN'
        end as rule_event
        from pg_rewrite r
        join pg_class c on r.ev_class = c.oid
        left join pg_namespace n on n.oid = c.relnamespace
        left join pg_description d on r.oid = d.objoid
        WHERE n.nspname NOT IN {self.exclude_schema}
            AND n.nspname NOT LIKE 'pg\_%'
        ORDER BY n.nspname, c.relname, r.rulename, rule_event
        """

        return self.__check_equals(query)

    @staticmethod
    def __normalize_constraint_definition(definition: str) -> str:
        """Normalize a constraint definition for comparison.

        PostgreSQL may represent functionally equivalent constraints differently,
        especially after dump/restore operations. This function normalizes common
        variations to enable accurate comparison.

        Args:
            definition: The constraint definition string from pg_get_constraintdef()

        Returns:
            Normalized constraint definition
        """
        if not definition:
            return definition

        # Normalize different ARRAY representations:
        # Before: (ARRAY['a'::type, 'b'::type])::type[] OR ARRAY[('a'::type)::text, ...]
        # After: Canonical form based on sorted elements

        # Strategy: Extract the constraint type and key values, ignoring formatting details
        # For ANY/ALL with arrays, extract just the operator and the array values

        # Remove extra parentheses around ARRAY expressions
        # (ARRAY[...])::type[] -> ARRAY[...]::type[]
        definition = re.sub(r"\(\(ARRAY\[(.*?)\]\)::(.*?)\[\]\)", r"ARRAY[\1]::\2[]", definition)

        # Also remove parentheses without cast: (ARRAY[...]) -> ARRAY[...]
        definition = re.sub(r"\(ARRAY\[([^\]]+)\]\)", r"ARRAY[\1]", definition)

        # Normalize array element casts: ('value'::type1)::type2 -> 'value'::type1
        # This handles the case where elements are double-cast
        definition = re.sub(r"\('([^']+)'::([^)]+)\)::(\w+)", r"'\1'::\2", definition)

        # Remove trailing array cast that may be present or absent: ::text[] or ::character varying[]
        # This is safe because the type information is already in each array element
        definition = re.sub(r"::(?:text|character varying)\[\]", "", definition)

        # Remove extra whitespace and normalize spacing
        definition = re.sub(r"\s+", " ", definition).strip()

        return definition

    def __check_equals(self, query, normalize_func=None) -> tuple[bool, list[DifferenceItem]]:
        """Check if the query results on the two databases are equals.

        Args:
            query: The SQL query to execute on both databases.
            normalize_func: Optional function to normalize specific fields in records.
                Should accept (dict, col_names) and return normalized dict.

        Returns:
            tuple: A tuple containing:
                - bool: True if the results are the same, False otherwise.
                - list[DifferenceItem]: A list of DifferenceItem objects with structured data.
        """
        self.cur1.execute(query)
        records1 = self.cur1.fetchall()

        self.cur2.execute(query)
        records2 = self.cur2.fetchall()

        result = True
        differences = []

        # Convert records to dictionaries based on column names
        col_names = [desc[0] for desc in self.cur1.description]

        # Create structured records
        structured1 = [dict(zip(col_names, record)) for record in records1]
        structured2 = [dict(zip(col_names, record)) for record in records2]

        # Apply normalization if provided
        if normalize_func:
            structured1 = [normalize_func(r, col_names) for r in structured1]
            structured2 = [normalize_func(r, col_names) for r in structured2]
            # Recreate records from normalized structured data
            records1 = [tuple(r[col] for col in col_names) for r in structured1]
            records2 = [tuple(r[col] for col in col_names) for r in structured2]

        # Create sets for comparison
        set1 = {str(tuple(r)) for r in records1}
        set2 = {str(tuple(r)) for r in records2}

        # Find differences
        removed = set1 - set2
        added = set2 - set1

        if removed or added:
            result = False

            # Map string representations back to structured data
            str_to_struct1 = {str(tuple(r)): s for r, s in zip(records1, structured1)}
            str_to_struct2 = {str(tuple(r)): s for r, s in zip(records2, structured2)}

            # Add removed items
            for item_str in removed:
                if item_str in str_to_struct1:
                    differences.append(
                        DifferenceItem(
                            type=DifferenceType.REMOVED,
                            content=str_to_struct1[item_str],
                        )
                    )

            # Add added items
            for item_str in added:
                if item_str in str_to_struct2:
                    differences.append(
                        DifferenceItem(
                            type=DifferenceType.ADDED,
                            content=str_to_struct2[item_str],
                        )
                    )

        return result, differences



# Base exception for all PUM errors
class PumException(Exception):
    """Base class for all exceptions raised by PUM.

    Version Added:
        1.0.0
    """


class PumDependencyError(PumException):
    """Exception when dependency are not resolved.

    Version Added:
        1.0.0
    """


# --- Configuration and Validation Errors ---


class PumConfigError(PumException):
    """Exception raised for errors in the PUM configuration.

    Version Added:
        1.0.0
    """


class PumInvalidChangelog(PumException):
    """Exception raised for invalid changelog.

    Version Added:
        1.0.0
    """


# --- Schema Migration Errors ---
class PumSchemaMigrationError(PumException):
    """Exception raised for errors related to schema migrations.

    Version Added:
        1.0.0
    """


class PumSchemaMigrationNoBaselineError(PumSchemaMigrationError):
    """Exception raised when no baseline version is found in the migration table.

    Version Added:
        1.0.0
    """


# --- Hook Errors ---


class PumHookError(PumException):
    """Exception raised for errors by an invalid hook.

    Version Added:
        1.0.0
    """


# --- Changelog/SQL Errors ---


class PumSqlError(PumException):
    """Exception raised for SQL-related errors in PUM.

    Version Added:
        1.0.0
    """


# --- Dump/Restore Errors (for dumper.py, if needed) ---


class PgDumpCommandError(PumException):
    """Exception raised for invalid pg_dump command.

    Version Added:
        1.0.0
    """


class PgDumpFailed(PumException):
    """Exception raised when pg_dump fails.

    Version Added:
        1.0.0
    """


class PgRestoreCommandError(PumException):
    """Exception raised for invalid pg_restore command.

    Version Added:
        1.0.0
    """


class PgRestoreFailed(PumException):
    """Exception raised when pg_restore fails.

    Version Added:
        1.0.0
    """



"""Main window for pgserviceparser GUI."""

from .compat import QtWidgets

QMainWindow = QtWidgets.QMainWindow

from .service_widget import PGServiceParserWidget


class _MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("PG Service Parser")
        self.setMinimumSize(600, 400)

        self._service_widget = PGServiceParserWidget()
        self.setCentralWidget(self._service_widget)



from enum import Enum


class ParameterType(Enum):
    """An enumeration of parameter types.
    This class defines the types of parameters that can be used in migration definitions.

    Attributes:
        BOOLEAN (str): Represents a boolean parameter type.
        INTEGER (str): Represents an integer parameter type.
        TEXT (str): Represents a text parameter type.
        DECIMAL (str): Represents a decimal parameter type.
        PATH (str): Represents a path parameter type.

    Version Added:
        1.0.0
    """

    BOOLEAN = "boolean"
    INTEGER = "integer"
    TEXT = "text"
    DECIMAL = "decimal"
    PATH = "path"

    def __str__(self) -> str:
        """Return the string value of the enum (e.g., 'integer' instead of 'ParameterType.INTEGER')."""
        return self.value


class ParameterDefinition:
    """A class to define a migration parameter.

    Version Added:
        1.0.0
    """

    def __init__(
        self,
        name: str,
        type: str | ParameterType,
        default: str | float | int | None = None,
        description: str | None = None,
        values: list | None = None,
        app_only: bool = False,
    ) -> None:
        """Initialize a ParameterDefintion instance.

        Args:
            name: The name of the parameter.
            type: The type of the parameter, as a string or ParameterType.
            default: The default value for the parameter. Defaults to None.
            description: A description of the parameter. Defaults to None.
            values: An optional list of allowed values for this parameter.
            app_only: If True, the parameter can be changed when recreating the app.
                Standard parameters (app_only=False) must remain the same across
                the whole application lifecycle. Defaults to False.

        Raises:
            ValueError: If type is a string and not a valid ParameterType.
            TypeError: If type is not a string or ParameterType.

        """
        self.name = name
        if isinstance(type, ParameterType):
            self.type = type
        elif isinstance(type, str):
            try:
                self.type = ParameterType(type)
            except ValueError:
                raise ValueError(f"Parameter '{name}' has an invalid type: {type}. ") from None
        else:
            raise TypeError("type must be a str or ParameterType")
        self.default = default
        self.description = description
        self.values = values
        self.app_only = app_only

    def __repr__(self) -> str:
        """Return a string representation of the ParameterDefinition instance."""
        return f"ParameterDefinition({self.name}, type: {self.type}, default: {self.default}, values: {self.values}, app_only: {self.app_only})"

    def __eq__(self, other: "ParameterDefinition") -> bool:
        """Check if two ParameterDefinition instances are equal."""
        if not isinstance(other, ParameterDefinition):
            return NotImplemented
        return (
            self.name == other.name
            and self.type == other.type
            and self.default == other.default
            and self.description == other.description
            and self.values == other.values
            and self.app_only == other.app_only
        )
